# bibfiles_db.py - load bibfiles into sqlite3, hash, assign ids (split/merge)

from __future__ import print_function

import logging
import difflib
import operator
import functools
import itertools
import contextlib
import collections

from six import string_types, iteritems, itervalues, viewkeys
from six.moves import map

import sqlalchemy as sa
import sqlalchemy.orm
import sqlalchemy.ext.declarative  # noqa: F401
from clldutils import path, dsv, jsonlib

from . import bibtex
from ..util import unique, group_first

__all__ = ['Database']

UNION_FIELDS = {'fn', 'asjp_name', 'isbn'}

IGNORE_FIELDS = {'crossref', 'numnote', 'glotto_id'}

log = logging.getLogger('pyglottolog')


class Database(object):
    """Bibfile collection parsed into an sqlite3 database."""

    @classmethod
    def from_bibfiles(cls, bibfiles, filepath, rebuild=False, page_size=32768):
        """If needed, (re)build the db from the bibfiles, hash, split/merge."""
        self = cls(filepath)

        if self.filepath.exists():
            if not rebuild and self.is_uptodate(bibfiles):
                return self
            path.remove(self.filepath)

        with self.engine.connect() as conn:
            if page_size is not None:
                conn.execute('PRAGMA page_size = %d' % page_size)
            Model.metadata.create_all(conn)

        with self.engine.connect() as conn:
            conn.execute('PRAGMA synchronous = OFF')
            conn.execute('PRAGMA journal_mode = MEMORY')
            conn = conn.execution_options(compiled_cache={})

            with conn.begin():
                import_bibfiles(conn, bibfiles)

            Entry.stats(conn)
            Value.fieldstats(conn)

            with conn.begin():
                generate_hashes(conn)

            Entry.hashstats(conn)
            Entry.hashidstats(conn)

            with conn.begin():
                assign_ids(conn)

        return self

    def __init__(self, filepath):
        if not isinstance(filepath, path.Path):
            filepath = path.Path(filepath)
        self.filepath = filepath
        self.engine = sa.create_engine('sqlite:///%s' % filepath, paramstyle='qmark')

    def is_uptodate(self, bibfiles, verbose=False):
        """Does the db have the same filenames, sizes, and mtimes as the given bibfiles?"""
        return File.same_as(self.engine, bibfiles, verbose=verbose)

    def stats(self, field_files=False):
        Entry.stats(self.engine)
        Value.fieldstats(self.engine, field_files)
        Entry.hashstats(self.engine)
        Entry.hashidstats(self.engine)

    def execute(self, statement, closing=True):
        cursor = self.engine.execute(statement)
        if closing:
            cursor = contextlib.closing(cursor)
        return cursor

    def to_bibfile(self, filepath, encoding='utf-8'):
        bibtex.save(self.merged(), str(filepath), sortkey=None, encoding=encoding)

    def to_csvfile(self, filename, encoding='utf-8', dialect='excel'):
        """Write a CSV file with one row for each entry in each bibfile."""
        select_rows = sa.select(
            [
                File.name.label('filename'), Entry.bibkey, Entry.hash,
                sa.cast(Entry.id, sa.Text).label('id'),
            ]).select_from(sa.join(File, Entry))\
            .order_by(sa.func.lower(File.name), sa.func.lower(Entry.bibkey), Entry.hash, Entry.id)
        with self.execute(select_rows) as cursor:
            with dsv.UnicodeWriter(filename, encoding=encoding, dialect=dialect) as writer:
                writer.writerow(cursor.keys())
                for row in cursor:
                    writer.writerow(row)

    def to_replacements(self, filename):
        """Write a JSON file with 301s from merged glottolog_ref_ids."""
        select_pairs = sa.select([Entry.refid.label('id'), Entry.id.label('replacement')])\
            .where(Entry.id != Entry.refid)\
            .order_by(Entry.id)
        with self.execute(select_pairs) as cursor:
            pairs = list(map(dict, cursor))
        with jsonlib.update(filename, default=[], indent=4) as repls:
            repls.extend(pairs)

    def trickle(self, bibfiles):
        """Write new/changed glottolog_ref_ids back into the given bibfiles."""
        assert Entry.allid(self.engine)
        if not self.is_uptodate(bibfiles, verbose=True):
            raise RuntimeError('trickle with an outdated db')  # pragma: no cover
        changed = (Entry.id != sa.func.coalesce(Entry.refid, -1))
        select_files = sa.select([File.pk, File.name])\
            .where(sa.exists().where(Entry.file_pk == File.pk).where(changed))\
            .order_by(File.name)
        select_changed = sa.select(
            [
                Entry.bibkey,
                sa.cast(Entry.refid, sa.Text).label('refid'),
                sa.cast(Entry.id, sa.Text).label('id'),
            ]).where(Entry.file_pk == sa.bindparam('file_pk')).where(changed)\
            .order_by(sa.func.lower(Entry.bibkey))
        with self.engine.connect() as conn:
            for file_pk, filename in conn.execute(select_files).fetchall():
                b = bibfiles[filename]
                entries = b.load()
                added = changed = 0
                for bibkey, refid, new in conn.execute(select_changed, file_pk=file_pk):
                    entrytype, fields = entries[bibkey]
                    old = fields.pop('glottolog_ref_id', None)
                    assert old == refid
                    if old is None:
                        added += 1
                    else:
                        changed += 1
                    fields['glottolog_ref_id'] = new
                print('%d changed %d added in %s' % (changed, added, b.id))
                b.save(entries)

    def merged(self):
        """Yield merged (bibkey, (entrytype, fields)) entries."""
        for (id, hash), grp in self:
            entrytype, fields = self._merged_entry(grp)
            fields['glottolog_ref_id'] = u'%d' % id
            yield hash, (entrytype, fields)

    def __iter__(self, chunksize=100):
        assert Entry.allid(self.engine)
        assert Entry.onetoone(self.engine)
        select_values = sa.select(
            [
                Entry.id, Entry.hash, Value.field, Value.value, File.name, Entry.bibkey,
            ]).select_from(sa.join(Entry, File).join(Value))\
            .where(sa.between(Entry.id, sa.bindparam('first'), sa.bindparam('last')))\
            .order_by(Entry.id, Value.field, File.priority.desc(), File.name, Entry.bibkey)
        get_id_hash, get_field = operator.itemgetter(0, 1), operator.itemgetter(2)
        with self.engine.connect() as conn:
            for first, last in Entry.windowed(conn, 'id', chunksize):
                cursor = conn.execute(select_values, first=first, last=last)
                for id_hash, grp in itertools.groupby(cursor, get_id_hash):
                    yield (
                        id_hash,
                        [
                            (field, [(vl, fn, bk) for _, _, _, vl, fn, bk in g])
                            for field, g in itertools.groupby(grp, get_field)])

    def __getitem__(self, key, _types=(tuple, int) + string_types):
        """Entry by (fn, bk) or merged entry by refid (old grouping) or hash (current grouping)."""
        if not isinstance(key, _types):
            raise ValueError  # pragma: no cover
        if isinstance(key, tuple):
            filename, bibkey = key
            entrytype, fields = self._entry(self.engine, filename, bibkey)
        else:
            grp = self._entrygrp(self.engine, key)
            entrytype, fields = self._merged_entry(grp)
        return key, (entrytype, fields)

    @staticmethod
    def _entry(bind, filename, bibkey):
        select_items = sa.select([Value.field, Value.value], bind=bind)\
            .select_from(sa.join(Value, Entry).join(File))\
            .where(File.name == filename)\
            .where(Entry.bibkey == bibkey)
        fields = dict(iter(select_items.execute()))
        if not fields:
            raise KeyError((filename, bibkey))
        return fields.pop('ENTRYTYPE'), fields

    @staticmethod
    def _merged_entry(grp, union=UNION_FIELDS, ignore=IGNORE_FIELDS, raw=False):
        # TODO: consider implementing (a subset of?) onlyifnot logic:
        # {'address': 'publisher', 'lgfamily': 'lgcode', 'publisher': 'school',
        # 'journal': 'booktitle'}
        fields = {
            field: values[0][0] if field not in union
            else ', '.join(unique(vl for vl, _, _ in values))
            for field, values in grp if field not in ignore}
        fields['src'] = ', '.join(sorted(set(
            fn.rpartition('.bib')[0] or fn for _, values in grp for _, fn, _ in values)))
        fields['srctrickle'] = ', '.join(sorted(set(
            '%s#%s' % (fn.rpartition('.bib')[0] or fn, bk)
            for _, values in grp for _, fn, bk in values)))
        if raw:
            return fields
        entrytype = fields.pop('ENTRYTYPE')
        return entrytype, fields

    @staticmethod
    def _entrygrp(bind, key, get_field=operator.itemgetter(0)):
        select_values = sa.select(
            [
                Value.field, Value.value, File.name, Entry.bibkey
            ],
            bind=bind).select_from(sa.join(Entry, File).join(Value))\
            .where((Entry.refid if isinstance(key, int) else Entry.hash) == key)\
            .order_by(Value.field, File.priority.desc(), File.name, Entry.bibkey)
        grouped = itertools.groupby(select_values.execute(), get_field)
        grp = [(field, [(vl, fn, bk) for _, vl, fn, bk in g]) for field, g in grouped]
        if not grp:
            raise KeyError(key)
        return grp

    def show_splits(self):
        other = sa.orm.aliased(Entry)
        select_entries = sa.select(
            [
                Entry.refid, Entry.hash, File.name.label('filename'), Entry.bibkey,
            ]).select_from(sa.join(Entry, File))\
            .order_by(Entry.refid, Entry.hash, File.name, Entry.bibkey)\
            .where(sa.exists().where(other.refid == Entry.refid).where(other.hash != Entry.hash))
        with self.engine.connect() as conn:
            for refid, group in group_first(conn.execute(select_entries)):
                self._print_group(conn, group)
                old = self._merged_entry(self._entrygrp(conn, refid), raw=True)
                cand = [
                    (hs, self._merged_entry(self._entrygrp(conn, hs), raw=True))
                    for hs in unique(hs for _, hs, _, _ in group)]
                new = min(cand, key=lambda p: distance(old, p[1]))[0]
                print('-> %s\n' % new)

    def show_merges(self):
        other = sa.orm.aliased(Entry)
        select_entries = sa.select(
            [
                Entry.hash, Entry.refid, File.name.label('filename'), Entry.bibkey,
            ]).select_from(sa.join(Entry, File))\
            .order_by(Entry.hash, Entry.refid.desc(), File.name, Entry.bibkey)\
            .where(sa.exists().where(other.hash == Entry.hash).where(other.refid != Entry.refid))
        with self.engine.connect() as conn:
            for hash, group in group_first(conn.execute(select_entries)):
                self._print_group(conn, group)
                new = self._merged_entry(self._entrygrp(conn, hash), raw=True)
                cand = [
                    (ri, self._merged_entry(self._entrygrp(conn, ri), raw=True))
                    for ri in unique(ri for _, ri, _, _ in group)]
                old = min(cand, key=lambda p: distance(new, p[1]))[0]
                print('-> %s\n' % old)

    @staticmethod
    def _print_group(conn, group, out=print):
        for row in group:
            out(row)
        for row in group:
            out('\t%r, %r, %r, %r' % Value.hashfields(conn, row['filename'], row['bibkey']))

    def _show(self, sql):
        with self.engine.connect() as conn:
            cursor = conn.execute(sql)
            for hash, group in group_first(cursor):
                self._print_group(conn, group)
                print()

    def show_identified(self):
        other = sa.orm.aliased(Entry)
        select_entries = sa.select(
            [
                Entry.hash, Entry.refid, File.name.label('filename'), Entry.bibkey,
            ]).select_from(sa.join(Entry, File))\
            .order_by(Entry.hash, Entry.refid != sa.null(), Entry.refid, File.name, Entry.bibkey)\
            .where(sa.exists().where(other.refid == sa.null()).where(other.hash == Entry.hash))\
            .where(sa.exists().where(other.refid != sa.null()).where(other.hash == Entry.hash))
        self._show(select_entries)

    def show_combined(self):
        other = sa.orm.aliased(Entry)
        select_entries = sa.select(
            [
                Entry.hash, File.name.label('filename'), Entry.bibkey,
            ]).select_from(sa.join(Entry, File))\
            .order_by(Entry.hash, File.name, Entry.bibkey)\
            .where(Entry.refid == sa.null())\
            .where(sa.exists().where(other.refid == sa.null()).where(other.hash == Entry.hash))
        self._show(select_entries)


Model = sa.ext.declarative.declarative_base()


class File(Model):

    __tablename__ = 'file'

    pk = sa.Column(sa.Integer, primary_key=True)
    name = sa.Column(sa.Text, nullable=False, unique=True)
    size = sa.Column(sa.Integer, nullable=False)
    mtime = sa.Column(sa.DateTime, nullable=False)
    priority = sa.Column(sa.Integer, nullable=False)

    @classmethod
    def same_as(cls, bind, bibfiles, verbose=False):
        ondisk = collections.OrderedDict(
            (b.fname.name, (b.size, b.mtime)) for b in bibfiles)
        select_files = sa.select([cls.name, cls.size, cls.mtime], bind=bind).order_by(cls.name)
        indb = collections.OrderedDict(
            (name, (size, mtime)) for name, size, mtime in select_files.execute())
        if dict(ondisk) == dict(indb):
            return True
        if verbose:
            ondisk_names, indb_names = (viewkeys(d) for d in (ondisk, indb))
            print('missing in db: %s' % list(ondisk_names - indb_names))
            print('missing on disk: %s' % list(indb_names - ondisk_names))
            print('differing in size/mtime: %s' % [
                o for o in (ondisk_names & indb_names) if ondisk[o] != indb[o]])
        return False


class Entry(Model):

    __tablename__ = 'entry'

    pk = sa.Column(sa.Integer, primary_key=True)
    file_pk = sa.Column(sa.ForeignKey('file.pk'), nullable=False)
    bibkey = sa.Column(sa.Text, nullable=False)
    # old glottolog_ref_id from bibfiles (previous hash groupings)
    refid = sa.Column(sa.Integer, index=True)
    # current groupings, m:n with refid (splits/merges):
    hash = sa.Column(sa.Text, index=True)
    # split-resolved refid (every srefid maps to exactly one hash):
    srefid = sa.Column(sa.Integer, index=True)
    # new glottolog_ref_id save to bibfiles (current hash groupings):
    id = sa.Column(sa.Integer, index=True)

    __table_args__ = (
        sa.UniqueConstraint(file_pk, bibkey),
    )

    @classmethod
    def allhash(cls, bind):
        return sa.select([~sa.exists().where(cls.hash == sa.null())], bind=bind).scalar()

    @classmethod
    def allid(cls, bind):
        return sa.select([~sa.exists().where(cls.id == sa.null())], bind=bind).scalar()

    @classmethod
    def onetoone(cls, bind):
        other = sa.orm.aliased(cls)
        return sa.select([~sa.exists().select_from(cls).where(sa.exists().where(sa.or_(
            sa.and_(other.hash == cls.hash, other.id != cls.id),
            sa.and_(other.id == cls.hash, other.hash != cls.hash))))], bind=bind).scalar()

    @classmethod
    def stats(cls, bind, out=log.info):
        out('entry stats:')
        select_n = sa.select([File.name.label('filename'), sa.func.count().label('n')], bind=bind)\
            .select_from(sa.join(cls, File)).group_by(cls.file_pk)
        out('\n'.join('%(filename)s %(n)d' % r for r in select_n.execute()))
        select_total = sa.select([sa.func.count()], bind=bind).select_from(cls)
        out('%d entries total' % select_total.scalar())

    @classmethod
    def hashstats(cls, bind, out=print):
        select_total = sa.select(
            [
                sa.func.count(cls.hash.distinct()).label('distinct'),
                sa.func.count(cls.hash).label('total'),
            ],
            bind=bind)
        tmpl = '%(distinct)d\tdistinct keyids (from %(total)d total)'
        out(tmpl % select_total.execute().first())

        sq1 = sa.select(
            [
                File.name.label('filename'),
                sa.func.count(cls.hash.distinct()).label('distinct'),
                sa.func.count(cls.hash).label('total'),
            ]).select_from(sa.join(cls, File))\
            .group_by(cls.file_pk).alias()
        other = sa.orm.aliased(cls)
        sq2 = sa.select(
            [
                File.name.label('filename'),
                sa.func.count(cls.hash.distinct()).label('unique'),
            ]).select_from(sa.join(cls, File))\
            .where(
                ~sa.exists()
                .where(other.hash == cls.hash)
                .where(other.file_pk != cls.file_pk))\
            .group_by(cls.file_pk).alias()
        select_files = sa.select(
            [
                sa.func.coalesce(sq2.c.unique, 0).label('unique'),
                sq1.c.filename, sq1.c.distinct, sq1.c.total,
            ],
            bind=bind)\
            .select_from(sq1.outerjoin(sq2, sq1.c.filename == sq2.c.filename))\
            .order_by(sq1.c.filename)
        tmpl = '%(unique)d\t%(filename)s (from %(distinct)d distinct of %(total)d total)'
        out('\n'.join(tmpl % r for r in select_files.execute()))

        select_multiple = sa.select([sa.func.count()], bind=bind)\
            .select_from(
                sa.select([1])
                .select_from(cls)
                .group_by(cls.hash)
                .having(sa.func.count(cls.file_pk.distinct()) > 1))
        out('%d\tin multiple files' % select_multiple.scalar())

    @classmethod
    def hashidstats(cls, bind, out=print):
        sq = sa.select([sa.func.count(cls.refid.distinct()).label('hash_nid')])\
            .where(cls.hash != sa.null())\
            .group_by(cls.hash)\
            .having(sa.func.count(cls.refid.distinct()) > 1).alias()
        select_nid = sa.select([sq.c.hash_nid, sa.func.count().label('n')], bind=bind)\
            .group_by(sq.c.hash_nid).order_by(sa.desc('n'))
        tmpl = '1 keyid %(hash_nid)d glottolog_ref_ids: %(n)d'
        out('\n'.join(tmpl % r for r in select_nid.execute()))

        sq = sa.select([sa.func.count(cls.hash.distinct()).label('id_nhash')])\
            .where(cls.refid != sa.null())\
            .group_by(cls.refid)\
            .having(sa.func.count(cls.hash.distinct()) > 1).alias()
        select_nhash = sa.select([sq.c.id_nhash, sa.func.count().label('n')], bind=bind)\
            .group_by(sq.c.id_nhash).order_by(sa.desc('n'))
        tmpl = '1 glottolog_ref_id %(id_nhash)d keyids: %(n)d'
        out('\n'.join(tmpl % r for r in select_nhash.execute()))

    @classmethod
    def windowed(cls, bind, colname, chunksize):
        col = cls.__table__.c[colname]
        select_col = sa.select([col.distinct()], bind=bind).order_by(col)
        with contextlib.closing(select_col.execute()) as cursor:
            for rows in iter(functools.partial(cursor.fetchmany, chunksize), []):
                (first,), (last,) = rows[0], rows[-1]
                yield first, last


class Value(Model):

    __tablename__ = 'value'

    entry_pk = sa.Column(sa.ForeignKey('entry.pk'), primary_key=True)
    field = sa.Column(sa.Text, primary_key=True)
    value = sa.Column(sa.Text, nullable=False)

    @classmethod
    def hashfields(cls, bind, filename, bibkey, _fields=('author', 'editor', 'year', 'title')):
        # also: extra_hash, volume (if not journal, booktitle, or series)
        select_items = sa.select([cls.field, cls.value], bind=bind)\
            .select_from(sa.join(Value, Entry).join(File))\
            .where(File.name == filename)\
            .where(Entry.bibkey == bibkey)\
            .where(cls.field.in_(_fields))
        fields = dict(iter(select_items.execute()))
        return tuple(fields.get(f) for f in _fields)

    @classmethod
    def fieldstats(cls, bind, with_files=False, out=print):
        select_n = sa.select([cls.field, sa.func.count().label('n')], bind=bind)\
            .group_by(cls.field).order_by(sa.desc('n'), cls.field)
        tmpl = '%(n)d\t%(field)s'
        if with_files:
            select_n = select_n.select_from(sa.join(Value, Entry).join(File))
            files = sa.func.replace(sa.func.group_concat(File.name.distinct()), ',', ', ')
            select_n.append_column(files.label('files'))
            tmpl += '\t%(files)s'
        out('\n'.join(tmpl % r for r in select_n.execute()))


def import_bibfiles(conn, bibfiles):
    log.info('importing bibfiles into a new db')
    assert conn.dialect.paramstyle == 'qmark'
    insert_file = sa.insert(File, bind=conn)\
        .compile(column_keys=['name', 'size', 'mtime', 'priority']).string
    insert_entry = sa.insert(Entry, bind=conn)\
        .compile(column_keys=['file_pk', 'bibkey', 'refid']).string
    insert_value = sa.insert(Value, bind=conn)\
        .compile(column_keys=['entry_pk', 'field', 'value']).string

    insert_file = functools.partial(conn.connection.execute, insert_file)
    insert_entry = functools.partial(conn.connection.execute, insert_entry)
    insert_values = functools.partial(conn.connection.executemany, insert_value)
    for b in bibfiles:
        file_pk = insert_file((b.fname.name, b.size, b.mtime, b.priority)).lastrowid
        for e in b.iterentries():
            bibkey = e.key
            entry_pk = insert_entry((file_pk, bibkey, e.fields.get('glottolog_ref_id'))).lastrowid
            fields = itertools.chain([('ENTRYTYPE', e.type)], iteritems(e.fields))
            insert_values(((entry_pk, field, value) for field, value in fields))


def generate_hashes(conn):
    from .libmonster import wrds, keyid

    words = collections.Counter()
    cursor = sa.select([Value.value], bind=conn).where(Value.field == 'title').execute()
    for rows in iter(functools.partial(cursor.fetchmany, 10000), []):
        for title, in rows:
            words.update(wrds(title))
    # TODO: consider dropping stop words/hapaxes from freq. distribution
    print('%d title words (from %d tokens)' % (len(words), sum(itervalues(words))))

    def windowed_entries(chunksize=500):
        select_files = sa.select([File.pk], bind=conn).order_by(File.name)
        select_bibkeys = sa.select([Entry.pk], bind=conn)\
            .where(Entry.file_pk == sa.bindparam('file_pk'))\
            .order_by(Entry.pk).execute
        for file_pk, in select_files.execute().fetchall():
            with contextlib.closing(select_bibkeys(file_pk=file_pk)) as cursor:
                for entry_pks in iter(functools.partial(cursor.fetchmany, chunksize), []):
                    (first,), (last,) = entry_pks[0], entry_pks[-1]
                    yield first, last

    select_bfv = sa.select([Entry.pk, Value.field, Value.value], bind=conn)\
        .select_from(sa.join(Value, Entry))\
        .where(Entry.pk.between(sa.bindparam('first'), sa.bindparam('last')))\
        .where(Value.field != 'ENTRYTYPE')\
        .order_by(Entry.pk).execute
    assert conn.dialect.paramstyle == 'qmark'
    update_entry = sa.update(Entry, bind=conn)\
        .values(hash=sa.bindparam('hash'))\
        .where(Entry.pk == sa.bindparam('entry_pk')).compile().string
    update_entry = functools.partial(conn.connection.executemany, update_entry)
    get_entry_pk = operator.itemgetter(0)
    for first, last in windowed_entries():
        rows = select_bfv(first=first, last=last)
        update_entry(
            ((keyid({k: v for _, k, v in grp}, words), entry_pk)
             for entry_pk, grp in itertools.groupby(rows, get_entry_pk)))


def assign_ids(conn, verbose=False):
    merged_entry, entrygrp = Database._merged_entry, Database._entrygrp
    other = sa.orm.aliased(Entry)

    assert Entry.allhash(conn)

    reset_entries = sa.update(Entry, bind=conn).values(id=sa.null(), srefid=Entry.refid)
    print('%d entries' % reset_entries.execute().rowcount)

    # resolve splits: srefid = refid only for entries from the most similar hash group
    nsplit = 0
    select_split = sa.select([Entry.refid, Entry.hash, File.name, Entry.bibkey], bind=conn)\
        .select_from(sa.join(Entry, File))\
        .order_by(Entry.refid, Entry.hash, File.name, Entry.bibkey)\
        .where(sa.exists().where(other.refid == Entry.refid).where(other.hash != Entry.hash))
    update_split = sa.update(Entry, bind=conn)\
        .where(Entry.refid == sa.bindparam('eq_refid'))\
        .where(Entry.hash != sa.bindparam('ne_hash'))\
        .values(srefid=sa.null()).execute
    for refid, group in group_first(select_split.execute()):
        old = merged_entry(entrygrp(conn, refid), raw=True)
        nsplit += len(group)
        cand = [
            (hs, merged_entry(entrygrp(conn, hs), raw=True))
            for hs in unique(hs for _, hs, _, _ in group)]
        new = min(cand, key=lambda p: distance(old, p[1]))[0]
        separated = update_split(eq_refid=refid, ne_hash=new).rowcount
        if verbose:
            for row in group:
                print(row)
            for _, _, fn, bk in group:
                print('\t%r, %r, %r, %r' % Value.hashfields(conn, fn, bk))
            print('-> %s' % new)
            print('%d: %d separated from %s\n' % (refid, separated, new))
    print('%d splitted' % nsplit)

    nosplits = sa.select(
        [
            ~sa.exists().select_from(Entry)
            .where(sa.exists().where(other.srefid == Entry.srefid).where(other.hash != Entry.hash))
        ],
        bind=conn)
    assert nosplits.scalar()

    # resolve merges: id = srefid of the most similar srefid group
    nmerge = 0
    select_merge = sa.select([Entry.hash, Entry.srefid, File.name, Entry.bibkey], bind=conn)\
        .select_from(sa.join(Entry, File))\
        .order_by(Entry.hash, Entry.srefid.desc(), File.name, Entry.bibkey)\
        .where(sa.exists().where(other.hash == Entry.hash).where(other.srefid != Entry.srefid))
    update_merge = sa.update(Entry, bind=conn)\
        .where(Entry.hash == sa.bindparam('eq_hash'))\
        .where(Entry.srefid != sa.bindparam('ne_srefid'))\
        .values(id=sa.bindparam('new_id')).execute
    for hash, group in group_first(select_merge.execute()):
        new = merged_entry(entrygrp(conn, hash), raw=True)
        nmerge += len(group)
        cand = [
            (ri, merged_entry(entrygrp(conn, ri), raw=True))
            for ri in unique(ri for _, ri, _, _ in group)]
        old = min(cand, key=lambda p: distance(new, p[1]))[0]
        merged = update_merge(eq_hash=hash, ne_srefid=old, new_id=old).rowcount
        if verbose:
            for row in group:
                print(row)
            for _, _, fn, bk in group:
                print('\t%r, %r, %r, %r' % Value.hashfields(conn, fn, bk))
            print('-> %s' % old)
            print('%s: %d merged into %d\n' % (hash, merged, old))
    print('%d merged' % nmerge)

    # unchanged entries
    update_unchanged = sa.update(Entry, bind=conn)\
        .where(Entry.id == sa.null())\
        .where(Entry.srefid != sa.null())\
        .values(id=Entry.srefid)
    print('%d unchanged' % update_unchanged.execute().rowcount)

    nomerges = sa.select(
        [
            ~sa.exists().select_from(Entry)
            .where(sa.exists().where(other.hash == Entry.hash).where(other.id != Entry.id))],
        bind=conn)
    assert nomerges.scalar()

    # identified
    update_identified = sa.update(Entry, bind=conn)\
        .where(Entry.refid == sa.null())\
        .where(
            sa.exists()
            .where(other.hash == Entry.hash)
            .where(other.id != sa.null()))\
        .values(
            id=sa.select([other.id]).where(other.hash == Entry.hash).where(other.id != sa.null()))
    print('%d identified (new/separated)' % update_identified.execute().rowcount)

    # assign new ids to hash groups of separated/new entries
    nextid = sa.select([sa.func.coalesce(sa.func.max(Entry.refid), 0) + 1], bind=conn).scalar()
    select_new = sa.select([Entry.hash], bind=conn)\
        .where(Entry.id == sa.null())\
        .group_by(Entry.hash)\
        .order_by(Entry.hash)
    assert conn.dialect.paramstyle == 'qmark'
    update_new = sa.update(Entry, bind=conn)\
        .values(id=sa.bindparam('new_id'))\
        .where(Entry.hash == sa.bindparam('eq_hash')).compile().string
    params = ((id, hash) for id, (hash,) in enumerate(select_new.execute(), nextid))
    dbapi_rowcount = conn.connection.executemany(update_new, params).rowcount
    # https://docs.python.org/2/library/sqlite3.html#sqlite3.Cursor.rowcount
    print('%d new ids (new/separated)' % (0 if dbapi_rowcount == -1 else dbapi_rowcount))

    assert Entry.allid(conn)
    assert Entry.onetoone(conn)

    # supersede relation
    select_superseded = sa.select([sa.func.count()], bind=conn).where(Entry.id != Entry.srefid)
    print('%d supersede pairs' % select_superseded.scalar())


def distance(left, right, weight={'author': 3, 'year': 3, 'title': 3, 'ENTRYTYPE': 2}):
    """Simple measure of the difference between two bibtex-field dicts."""
    if not (left or right):
        return 0.0

    keys = viewkeys(left) & viewkeys(right)
    if not keys:
        return 1.0

    weights = {k: weight.get(k, 1) for k in keys}
    ratios = (
        w * difflib.SequenceMatcher(None, left[k], right[k]).ratio()
        for k, w in iteritems(weights))
    return 1 - (sum(ratios) / sum(itervalues(weights)))



# bibfiles.py - ordered collection of bibfiles with load/save api

from __future__ import unicode_literals, print_function

import re
import collections
import unicodedata
import datetime

from six import string_types

import attr

from clldutils.misc import UnicodeMixin
from clldutils.path import memorymapped, Path
from clldutils.source import Source
from clldutils.text import split_text
from clldutils.inifile import INI

from . import bibtex
from .bibfiles_db import Database

__all__ = ['BibFiles', 'BibFile', 'Entry']

BIBFILES = 'bibfiles.sqlite3'


class BibFiles(list):
    """Ordered collection of `BibFile` objects accessible by filname or index."""

    @classmethod
    def from_path(cls, path):
        """BibTeX files from `<path>/bibtex/*.bib` if listed in `<path>/BIBFILES.ini`."""
        if not isinstance(path, Path):
            path = Path(path)
        ini = INI.from_file(path / 'BIBFILES.ini', interpolation=None)
        return cls(cls._iterbibfiles(ini, path / 'bibtex'))

    @staticmethod
    def _iterbibfiles(ini, bibtex_path):
        for sec in ini.sections():
            if sec.endswith('.bib'):
                fpath = bibtex_path / sec
                if not fpath.exists():  # pragma: no cover
                    raise ValueError('invalid bibtex file referenced in BIBFILES.ini')
                yield BibFile(fname=fpath, **ini[sec])

    def __init__(self, bibfiles):
        super(BibFiles, self).__init__(bibfiles)
        self._map = {b.fname.name: b for b in self}

    def __getitem__(self, index_or_filename):
        """Retrieve a bibfile by index or filename."""
        if isinstance(index_or_filename, string_types):
            return self._map[index_or_filename]
        return super(BibFiles, self).__getitem__(index_or_filename)

    def to_sqlite(self, filepath=BIBFILES, rebuild=False):
        """Return a database with the bibfiles loaded."""
        return Database.from_bibfiles(self, filepath, rebuild=rebuild)

    def roundtrip_all(self):
        """Load and save all bibfiles with the current settings."""
        return [b.roundtrip() for b in self]


def file_if_exists(i, a, value):
    if value.exists() and not value.is_file():
        raise ValueError('invalid path')  # pragma: no cover


@attr.s
class BibFile(UnicodeMixin):

    fname = attr.ib(validator=file_if_exists)
    name = attr.ib(default=None)
    title = attr.ib(default=None)
    description = attr.ib(default=None)
    abbr = attr.ib(default=None)
    encoding = attr.ib(default='utf-8')
    normalize = attr.ib(default='NFC')
    sortkey = attr.ib(
        default=None,
        converter=lambda s: None if s is None or s.lower() == 'none' else s)
    priority = attr.ib(default=0, converter=int)
    url = attr.ib(default=None)

    @property
    def id(self):
        return self.fname.stem

    def __getitem__(self, item):
        if item.startswith(self.id + ':'):
            item = item.split(':', 1)[1]
        text = None
        with memorymapped(self.fname) as string:
            m = re.search(b'@[A-Za-z]+\{' + re.escape(item.encode(self.encoding)), string)
            if m:
                next = string.find(b'\n@', m.end())
                if next >= 0:
                    text = string[m.start():next]
                else:
                    text = string[m.start():]
        if text:
            for k, (t, f) in bibtex.iterentries_from_text(text, encoding=self.encoding):
                return Entry(k, t, f, self)
        raise KeyError(item)

    def visit(self, visitor=None):
        entries = collections.OrderedDict()
        for entry in self.iterentries():
            if visitor is None or visitor(entry) is not True:
                entries[entry.key] = (entry.type, entry.fields)
        self.save(entries)

    @property
    def size(self):
        return self.fname.stat().st_size

    @property
    def mtime(self):
        return datetime.datetime.fromtimestamp(self.fname.stat().st_mtime)

    def iterentries(self):
        for k, (t, f) in bibtex.iterentries(filename=self.fname, encoding=self.encoding):
            yield Entry(k, t, f, self)

    def keys(self):
        return ['{0}:{1}'.format(self.id, e.key) for e in self.iterentries()]

    @property
    def glottolog_ref_id_map(self):
        return {
            e.key: e.fields['glottolog_ref_id'] for e in self.iterentries()
            if 'glottolog_ref_id' in e.fields}

    def update(self, fname, log=None):
        entries, new = collections.OrderedDict(), 0
        ref_id_map = self.glottolog_ref_id_map
        for key, (type_, fields) in bibtex.iterentries(fname, self.encoding):
            if key in ref_id_map and 'glottolog_ref_id' not in fields:
                fields['glottolog_ref_id'] = ref_id_map[key]
            else:
                new += 1
            entries[key] = (type_, fields)
        self.save(entries)
        if log:
            log.info('{0} new entries'.format(new))

    def load(self, preserve_order=None):
        """Return entries as bibkey -> (entrytype, fields) dict."""
        if preserve_order is None:
            preserve_order = self.sortkey is None
        return bibtex.load(self.fname, preserve_order, encoding=self.encoding)

    def save(self, entries):
        """Write bibkey -> (entrytype, fields) map to file."""
        bibtex.save(
            entries,
            filename=self.fname,
            sortkey=self.sortkey,
            encoding=self.encoding,
            normalize=self.normalize)

    def __unicode__(self):
        return '<%s %s>' % (self.__class__.__name__, self.fname.name)

    def check(self, log):
        entries = self.load()  # bare BibTeX syntax
        invalid = bibtex.check(filename=self.fname)  # names/macros etc.
        verdict = ('(%d invalid)' % invalid) if invalid else 'OK'
        method = log.warn if invalid else log.info
        method('%s %d %s' % (self, len(entries), verdict))
        return len(entries), verdict

    def roundtrip(self):
        print(self)
        self.save(self.load())

    def show_characters(self, include_plain=False):
        """Display character-frequencies (excluding printable ASCII)."""
        with self.fname.open(encoding=self.encoding) as fd:
            text = fd.read()
        hist = collections.Counter(text)
        table = '\n'.join(
            '%d\t%-9r\t%s\t%s' % (n, c, c, unicodedata.name(c, ''))
            for c, n in hist.most_common()
            if include_plain or not 20 <= ord(c) <= 126)
        print(table)


@attr.s
class Entry(UnicodeMixin):

    key = attr.ib()
    type = attr.ib()
    fields = attr.ib()
    bib = attr.ib()

    # FIXME: add method to apply triggers!

    lgcode_regex = '[a-z0-9]{4}[0-9]{4}|[a-z]{3}|NOCODE_[A-Z][^\s\]]+'
    lgcode_in_brackets_pattern = re.compile("\[(" + lgcode_regex + ")\]")
    recomma = re.compile("[,/]\s?")
    lgcode_pattern = re.compile(lgcode_regex + "$")

    def __unicode__(self):
        """
        :return: BibTeX representation of the entry.
        """
        res = "@%s{%s" % (self.type, self.key)
        for k, v in bibtex.fieldorder.itersorted(self.fields):
            res += ',\n    %s = {%s}' % (k, v.strip() if hasattr(v, 'strip') else v)
        res += '\n}\n' if self.fields else ',\n}\n'
        return res

    def text(self):
        """
        :return: Text linearization of the entry.
        """
        return Source(self.type, self.key, _check_id=False, **self.fields).text()

    @property
    def id(self):
        return '{0}:{1}'.format(self.bib.id, self.key)

    @classmethod
    def lgcodes(cls, string):
        if string is None:
            return []
        codes = cls.lgcode_in_brackets_pattern.findall(string)
        if not codes:
            # ... or as comma separated list of identifiers.
            parts = [p.strip() for p in cls.recomma.split(string)]
            codes = [p for p in parts if cls.lgcode_pattern.match(p)]
            if len(codes) != len(parts):
                codes = []
        return codes

    @staticmethod
    def parse_ca(s):
        if s:
            match = re.search('computerized assignment from "(?P<trigger>[^\"]+)"', s)
            if match:
                return match.group('trigger')

    def languoids(self, langs_by_codes):
        res = []
        if 'lgcode' in self.fields:
            for code in self.lgcodes(self.fields['lgcode']):
                if code in langs_by_codes:
                    res.append(langs_by_codes[code])
        return res, self.parse_ca(self.fields.get('lgcode'))

    def doctypes(self, hhtypes):
        res = []
        if 'hhtype' in self.fields:
            for ss in split_text(self.fields['hhtype'], separators=',;'):
                ss = ss.split('(')[0].strip()
                if ss in hhtypes:
                    res.append(hhtypes[ss])
        return res, self.parse_ca(self.fields.get('hhtype'))



# coding: utf8
from __future__ import unicode_literals, print_function, division
from collections import defaultdict, Counter
from itertools import chain
import os
import sys
import re
import argparse
import subprocess
from json import dumps
from string import Template

from termcolor import colored
from csvw import TableGroup, Column
import pycldf.util
from clldutils.jsonlib import load
from clldutils.clilib import command, ParserError
from clldutils.misc import slug
from clldutils.markup import Table
from clldutils.path import Path, write_text, read_text, git_describe

import pyglottolog
import pyglottolog.iso
from .languoids import Languoid, Level, Reference
from . import fts
from . import lff
from .monster import compile
from .references import evobib
from .util import message, sprint


@command()
def languoids(args):
    """
    glottolog languoids [--output=OUTDIR] [--version=VERSION]
    """
    parser = argparse.ArgumentParser()
    parser.add_argument(
        '--output', help='An existing directory for the output', default='.', type=Path)
    parser.add_argument(
        '--version', help='Version string if it cannot be inferred', default=args.repos.describe())
    xargs = parser.parse_args(args.args)

    out = xargs.output / 'glottolog-languoids-{0}.csv'.format(xargs.version)
    md = xargs.output / (out.name + '-metadata.json')
    tg = TableGroup.fromvalue({
        "@context": "http://www.w3.org/ns/csvw",
        "dc:version": xargs.version,
        "dc:": "Harald Hammarström, Robert Forkel & Martin Haspelmath. "
               "clld/glottolog: Glottolog database (Version {0}) [Data set]. "
               "Zenodo. http://doi.org/10.5281/zenodo.596479".format(xargs.version),
        "tables": [
            load(pycldf.util.pkg_path('components', 'LanguageTable-metadata.json'))
    ]})
    tg.tables[0].url = out.name
    for col in [
        dict(name='Classification', separator='/'),
        dict(name='Family_Glottocode'),
        dict(name='Family_Name'),
        dict(name='Language_Glottocode'),
        dict(name='Language_Name'),
        dict(name='Level', datatype=dict(base='string', format='family|language|dialect')),
        dict(name='Status'),
    ]:
        tg.tables[0].tableSchema.columns.append(Column.fromvalue(col))

    langs = []
    for lang in args.repos.languoids():
        lid, lname = None, None
        if lang.level == Level.language:
            lid, lname = lang.id, lang.name
        elif lang.level == Level.dialect:
            for lname, lid, level in reversed(lang.lineage):
                if level == Level.language:
                    break
            else:  # pragma: no cover
                raise ValueError
        langs.append(dict(
            ID=lang.id,
            Name=lang.name,
            Macroarea=lang.macroareas[0] if lang.macroareas else None,
            Latitude=lang.latitude,
            Longitude=lang.longitude,
            Glottocode=lang.id,
            ISO639P3code=lang.iso,
            Classification=[c[1] for c in lang.lineage],
            Language_Glottocode=lid,
            Language_Name=lname,
            Family_Name=lang.lineage[0][0] if lang.lineage else None,
            Family_Glottocode=lang.lineage[0][1] if lang.lineage else None,
            Level=lang.level.name,
            Status=lang.endangerment.description if lang.endangerment else None,
        ))

    tg.to_file(md)
    print('Metadata: {0}'.format(md))
    tg.tables[0].write(langs, fname=out)
    print('Data: {0}'.format(out))


@command()
def htmlmap(args, min_langs_for_legend_item=10):
    """
    glottolog htmlmap [OUTDIR]
    """
    nodes = {n.id: n for n in args.repos.languoids()}
    legend = Counter()

    langs = []
    for n in nodes.values():
        if n.level == Level.language and n.latitude != None:
            fid = n.lineage[0][1] if n.lineage else n.id
            if (not nodes[fid].category.startswith('Pseudo')) or fid == n.id:
                langs.append((n, fid))
                legend.update([fid])

    color_map = {fid: "{0:0{1}X}".format((i + 1) * 10, 3)
                 for i, fid in enumerate(sorted(legend.keys()))}

    def l2f(t):
        n, fid = t
        lon, lat = n.longitude, n.latitude
        if lon <= -26:
            lon += 360  # make the map pacific-centered.

        return {
            "geometry": {"coordinates": [lon, lat], "type": "Point"},
            "id": n.id,
            "properties": {
                "name": n.name,
                "color": color_map[fid],
                "family": nodes[fid].name,
                "family_id": fid,
            },
            "type": "Feature"
        }

    def legend_item(fid, c):
        return \
            '<span style="background-color: #{0}; border: 1px solid black;">'\
            '&nbsp;&nbsp;&nbsp;</span> '\
            '<a href="http://glottolog.org/resource/languoid/id/{1}">{2}</a> ({3})'.format(
                color_map[fid], fid, nodes[fid].name, c)

    geojson = {
        "features": list(map(l2f, langs)),
        "properties": {
            "legend": {
                fid: legend_item(fid, c) for fid, c in legend.most_common() if
                c >= min_langs_for_legend_item},
        },
        "type": "FeatureCollection"
    }

    def rendered_template(name, **kw):
        return Template(read_text(
            Path(pyglottolog.__file__).parent.joinpath('templates', 'htmlmap', name))
        ).substitute(**kw)

    jsname = 'glottolog_map.json'
    outdir = Path('.') if not args.args else Path(args.args[0])
    write_text(
        outdir.joinpath(jsname),
        rendered_template('htmlmap.js', geojson=dumps(geojson, indent=4)))
    html = outdir.joinpath('glottolog_map.html')
    write_text(
        html,
        rendered_template(
            'htmlmap.html',
            version=git_describe(args.repos.repos),
            jsname=jsname,
            nlangs=len(langs)))
    print(html.resolve().as_uri())


@command()
def iso2codes(args):
    """
    Map ISO codes to the list of all Glottolog languages and dialects subsumed "under" it.
    """
    from clldutils.dsv import UnicodeWriter

    nodes = list(args.repos.languoids())

    res = {}
    for node in nodes:
        if node.iso:
            res[node.id] = (node.iso, set())

    for node in nodes:
        if node.level == Level.family or node.id in res:
            continue
        for nid in res:
            matched = False
            for l in node.lineage:
                if l[1] == nid:
                    res[nid][1].add(node.id)
                    matched = True
                    break
            if matched:
                break

    outdir = Path('.') if not args.args else Path(args.args[0])
    with UnicodeWriter(outdir / 'iso2glottocodes.csv') as writer:
        writer.writerow(['iso', 'glottocodes'])
        for gc, (iso, gcs) in res.items():
            writer.writerow([iso, ';'.join([gc] + list(gcs))])


@command('evobib')
def _evobib(args):
    evobib.download(args.repos.bibfiles['evobib.bib'], args.log)


@command()
def roundtrip(args):
    """Load/save the bibfile with the given name."""
    args.repos.bibfiles[args.args[0]].roundtrip()


@command()
def bibfiles_db(args):
    """(Re-)create bibfiles sqlite3 database in the current directory."""
    args.repos.bibfiles.to_sqlite(rebuild=True)


@command()
def copy_benjamins(args, name='benjamins.bib'):  # pragma: no cover
    """
    glottolog copy_benjamins /path/to/benjamins/benjamins.bib
    """
    args.repos.bibfiles[name].update(args.args[0], log=args.log)


@command()
def isobib(args):  # pragma: no cover
    """Update iso6393.bib - the file of references for ISO 639-3 change requests."""
    pyglottolog.iso.bibtex(args.repos, args.log)


@command()
def isoretirements(args):  # pragma: no cover
    """Update retirement info in language info files."""
    pyglottolog.iso.retirements(args.repos, args.log)


def existing_lang(args):
    if not args.args:
        raise ParserError('No languoid specified')
    lang = args.repos.languoid(args.args[0])
    if not lang:
        raise ParserError('Invalid languoid spec')
    return lang


@command()
def show(args):
    """Display details of a Glottolog object.

    glottolog show <GLOTTOCODE>|<ISO-CODE>|<BIBTEXKEY>
    """
    if args.args and ':' in args.args[0]:
        if args.args[0].startswith('**'):
            ref = Reference.from_string(args.args[0])
        else:
            ref = Reference(key=args.args[0])
        sprint('Glottolog reference {0}'.format(ref), attrs=['bold', 'underline'])
        print()
        src = ref.get_source(args.repos)
        sprint(src.text())
        print()
        sprint(src)
        return
    lang = existing_lang(args)
    print()
    sprint('Glottolog languoid {0}'.format(lang.id), attrs=['bold', 'underline'])
    print()
    sprint('Classification:', attrs=['bold', 'underline'])
    args.repos.ascii_tree(lang, maxlevel=1)
    print()
    sprint('Info:', attrs=['bold', 'underline'])
    sprint('Path: {0}'.format(lang.fname), 'green', attrs=['bold'])
    sources = lang.sources
    if sources:
        del lang.cfg['sources']['glottolog']
        del lang.cfg['sources']
    for line in lang.cfg.write_string().split('\n'):
        if not line.startswith('#'):
            sprint(line, None, attrs=['bold'] if line.startswith('[') else [])
    sprint('Sources:', attrs=['bold', 'underline'])
    for src in sources:
        src = src.get_source(args.repos)
        sprint(src.id, color='green')
        sprint(src.text())
        print()


@command()
def edit(args):
    """Open a languoid's INI file in a text editor.

    glottolog edit <GLOTTOCODE>|<ISO-CODE>
    """
    lang = existing_lang(args)
    if sys.platform.startswith('os2'):  # pragma: no cover
        cmd = ['open']
    elif sys.platform.startswith('linux'):
        cmd = ['xdg-open']
    elif sys.platform.startswith('win'):  # pragma: no cover
        cmd = []
    else:  # pragma: no cover
        print(lang.fname)
        return
    cmd.append(lang.fname.as_posix())
    subprocess.call(cmd)


@command()
def create(args):
    """Create a new languoid directory for a languoid specified by name and level.

    glottolog create <parent> <name> <level>
    """
    assert args.args[2] in ['family', 'language', 'dialect']
    parent = args.repos.languoid(args.args[0]) or None
    outdir = parent.dir if parent else args.repos.tree
    lang = Languoid.from_name_id_level(
        outdir,
        args.args[1],
        args.repos.glottocodes.new(args.args[1]),
        getattr(Level, args.args[2]),
        **dict(prop.split('=') for prop in args.args[3:]))

    print("Info written to %s" % lang.write_info(outdir=outdir))


@command()
def bib(args):
    """Compile the monster bibfile from the BibTeX files listed in references/BIBFILES.ini

    glottolog bib [rebuild]
    """
    compile(args.repos, args.log, rebuild=bool(args.args))


@command()
def tree(args):
    """Print the classification tree starting at a specific languoid.

    glottolog tree <GLOTTOCODE>|<ISO-CODE> [MAXLEVEL]

    MAXLEVEL [family|language|dialect] will limit the displayed children.
    """
    start = existing_lang(args)
    maxlevel = None
    if len(args.args) > 1:
        try:
            maxlevel = int(args.args[1])
        except Exception:
            maxlevel = getattr(Level, args.args[1], None)
    args.repos.ascii_tree(start, maxlevel=maxlevel)


@command(usage="""
Print the classification tree starting at a specific languoid in Newick format.

    glottolog newick [--template="{{l.id}}"] [<GLOTTOCODE>|<ISO-CODE>]

The --template option can be used to control the node labels in the Newick string.
Values for this option must be valid python format strings expecting a single
template variable `l` which points to the Languoid instance.
In addition to Languoid attributes and properties specified as "{{l.<attr>}}",
e.g. "{{l.id}}" for the Glottocode of a Languoid, the following custom format specs
can be used:
{0}""".format(
    '\n'.join('    l:{0}\t{1[1]}'.format(k, v) for k, v in Languoid._format_specs.items())))
def newick(args):
    parser = argparse.ArgumentParser(prog='newick')
    parser.add_argument('root', nargs='?', default=None, help='root node')
    parser.add_argument('--template', help='node label template', default=None)
    xargs = parser.parse_args(args.args)
    sprint(args.repos.newick_tree(xargs.root, template=xargs.template))


@command()
def index(args):
    """Create an index page listing and linking to all languoids of a specified level.

    glottolog index (family|language|dialect|all)
    """
    def make_index(level, languoids, repos):
        fname = dict(
            language='languages', family='families', dialect='dialects')[level.name]
        links = defaultdict(dict)
        for lang in languoids:
            label = '{0.name} [{0.id}]'.format(lang)
            if lang.iso:
                label += '[%s]' % lang.iso
            links[slug(lang.name)[0]][label] = \
                lang.fname.relative_to(repos.languoids_path())

        with repos.languoids_path(fname + '.md').open('w', encoding='utf8') as fp:
            fp.write('## %s\n\n' % fname.capitalize())
            fp.write(' '.join(
                '[-%s-](%s_%s.md)' % (i.upper(), fname, i) for i in sorted(links.keys())))
            fp.write('\n')

        for i, langs in links.items():
            with repos.languoids_path(
                    '%s_%s.md' % (fname, i)).open('w', encoding='utf8') as fp:
                for label in sorted(langs.keys()):
                    fp.write('- [%s](%s)\n' % (label, langs[label]))

    langs = list(args.repos.languoids())
    for level in Level:
        if not args.args or args.args[0] == level.name:
            make_index(level, [l for l in langs if l.level == level], args.repos)


@command()
def check(args):
    """Check the glottolog data for consistency.

    glottolog check [tree|refs]
    """
    def error(obj, msg):
        args.log.error(message(obj, msg))

    def warn(obj, msg):
        args.log.warn(message(obj, msg))

    def info(obj, msg):
        args.log.info(message(obj, msg))

    what = args.args[0] if args.args else 'all'

    if what in ['all', 'refs']:
        for bibfile in args.repos.bibfiles:
            bibfile.check(args.log)

    if what not in ['all', 'tree']:
        return

    hhkeys = args.repos.bibfiles['hh.bib'].keys()
    iso = args.repos.iso
    args.log.info('checking ISO codes against %s' % iso)
    args.log.info('checking tree at %s' % args.repos)
    by_level = Counter()
    by_category = Counter()
    iso_in_gl, languoids, iso_splits, hid = {}, {}, [], {}
    names = defaultdict(set)

    for lang in args.repos.languoids():
        # duplicate glottocodes:
        if lang.id in languoids:
            error(
                lang.id,
                'duplicate glottocode\n{0}\n{1}'.format(languoids[lang.id].dir, lang.dir))
        languoids[lang.id] = lang

    for lang in languoids.values():
        ancestors = lang.ancestors_from_nodemap(languoids)
        children = lang.children_from_nodemap(languoids)

        if lang.latitude and not (-90 <= lang.latitude <= 90):
            error(lang, 'invalid latitude: {0}'.format(lang.latitude))
        if lang.longitude and not (-180 <= lang.longitude <= 180):
            error(lang, 'invalid longitude: {0}'.format(lang.longitude))

        assert isinstance(lang.countries, list)
        assert isinstance(lang.macroareas, list)

        if 'sources' in lang.cfg:
            for ref in Reference.from_list(lang.cfg.getlist('sources', 'glottolog')):
                if ref.provider == 'hh' and ref.key not in hhkeys:
                    error(lang, 'missing source: {0}'.format(ref))

        for attr in ['classification_comment', 'ethnologue_comment']:
            obj = getattr(lang, attr)
            if obj:
                obj.check(lang, hhkeys, args.log)

        names[lang.name].add(lang)
        by_level.update([lang.level.name])
        if lang.level == Level.language:
            by_category.update([lang.category])

        if iso and lang.iso:
            if lang.iso not in iso:
                warn(lang, 'invalid ISO-639-3 code [%s]' % lang.iso)
            else:
                isocode = iso[lang.iso]
                if lang.iso in iso_in_gl:
                    error(isocode,
                          'duplicate: {0}, {1}'.format(iso_in_gl[lang.iso].id, lang.id))
                iso_in_gl[lang.iso] = lang
                if isocode.is_retired and lang.category != 'Bookkeeping':
                    if isocode.type == 'Retirement/split':
                        iso_splits.append(lang)
                    else:
                        msg = repr(isocode)
                        level = info
                        if len(isocode.change_to) == 1:
                            level = warn
                            msg += ' changed to [%s]' % isocode.change_to[0].code
                        level(lang, msg)

        if lang.hid is not None:
            if lang.hid in hid:
                error(
                    lang.hid,
                    'duplicate hid\n{0}\n{1}'.format(languoids[hid[lang.hid]].dir, lang.dir))
            else:
                hid[lang.hid] = lang.id

        if not lang.id.startswith('unun9') and lang.id not in args.repos.glottocodes:
            error(lang, 'unregistered glottocode')
        for attr in ['level', 'name']:
            if not getattr(lang, attr):
                error(lang, 'missing %s' % attr)
        if lang.level == Level.language:
            parent = ancestors[-1] if ancestors else None
            if parent and parent.level != Level.family:
                error(lang, 'invalid nesting of language under {0}'.format(parent.level))
            for child in children:
                if child.level != Level.dialect:
                    error(child,
                          'invalid nesting of {0} under language'.format(child.level))
        elif lang.level == Level.family:
            for d in lang.dir.iterdir():
                if d.is_dir():
                    break
            else:
                error(lang, 'family without children')

    if iso:
        changed_to = set(chain(*[code.change_to for code in iso.retirements]))
        for code in sorted(iso.languages):
            if code.type == 'Individual/Living':
                if code not in changed_to:
                    if code.code not in iso_in_gl:
                        info(repr(code), 'missing')
        for lang in iso_splits:
            isocode = iso[lang.iso]
            missing = [s.code for s in isocode.change_to if s.code not in iso_in_gl]
            if missing:
                warn(lang, '{0} missing new codes: {1}'.format(
                    repr(isocode), ', '.join(missing)))

    for name, gcs in sorted(names.items()):
        if len(gcs) > 1:
            # duplicate names:
            method = error
            if len([1 for n in gcs if n.level != Level.dialect]) <= 1:
                # at most one of the languoids is not a dialect, just warn
                method = warn
            if len([1 for n in gcs
                    if (not n.lineage) or (n.lineage[0][1] != 'book1242')]) <= 1:
                # at most one of the languoids is not in bookkeping, just warn
                method = warn
            method(name, 'duplicate name: {0}'.format(', '.join(sorted(
                ['{0} <{1}>'.format(n.id, n.level.name[0]) for n in gcs]))))

    def log_counter(counter, name):
        msg = [name + ':']
        maxl = max([len(k) for k in counter.keys()]) + 1
        for k, l in counter.most_common():
            msg.append(('{0:<%s} {1:>8,}' % maxl).format(k + ':', l))
        msg.append(('{0:<%s} {1:>8,}' % maxl).format('', sum(list(counter.values()))))
        print('\n'.join(msg))

    log_counter(by_level, 'Languoids by level')
    log_counter(by_category, 'Languages by category')
    return by_level


@command()
def metadata(args):
    """List all metadata fields used in languoid INI files and their frequency.

    glottolog metadata
    """
    ops = defaultdict(Counter)

    for l in args.repos.languoids():
        for secname, sec in l.cfg.items():
            ops[secname].update(opt for opt, val in sec.items() if val)

    ops.pop('DEFAULT', None)

    t = Table('section', 'option', 'count')
    for section, options in ops.items():
        t.append([section, '', float(sum(options.values()))])
        for k, n in options.most_common():
            t.append(['', k, float(n)])
    print(t.render(condensed=False, floatfmt=',.0f'))


@command()
def refsearch(args):
    """Search Glottolog references

    glottolog refsearch "QUERY"

    E.g.:
    - glottolog refsearch "Izi provider:hh"
    - glottolog refsearch "author:Haspelmath provider:wals"
    """
    count, results = fts.search(args.repos, args.args[0])
    table = Table('ID', 'Author', 'Year', 'Title')
    for res in results:
        table.append([res.id, res.author, res.year, res.title])
    sprint(table.render(tablefmt='simple'))
    print('({} matches)'.format(count))


@command()
def refindex(args):
    """Index all bib files for use with `glottolog refsearch`.

    glottolog refindex

    This will take about 15 minutes and create an index of about 450 MB.
    """
    return fts.build_index(args.repos, args.log)


@command()
def langsearch(args):
    """Search Glottolog languoids

    glottolog langsearch "QUERY"
    """
    def highlight(text):
        res, i = '', 0
        for m in re.finditer('\[\[(?P<m>[^\]]+)\]\]', text):
            res += text[i:m.start()]
            res += colored(m.group('m'), 'red', attrs=['bold'])
            i = m.end()
        res += text[i:]
        return res + '\n'

    count, results = fts.search_langs(args.repos, args.args[0])
    cwd = os.getcwd()
    print('{} matches'.format(count))
    for res in results:
        try:
            p = Path(res.fname).relative_to(Path(cwd))
        except ValueError:
            p = res.fname
        sprint('{0.name} [{0.id}] {0.level}'.format(res), color=None, attrs=['bold'])
        sprint(p, color='green')
        sprint(highlight(res.highlights) if res.highlights else '')
    print('{} matches'.format(count))


@command()
def langindex(args):
    """Index all bib files for use with `glottolog langsearch`.

    glottolog langindex

    This will take a couple of minutes and create an index of about 60 MB.
    """
    return fts.build_langs_index(args.repos, args.log)


@command()
def tree2lff(args):
    """Create lff.txt and dff.txt from the current languoid tree.

    glottolog tree2lff
    """
    lff.tree2lff(args.repos, args.log)


@command()
def lff2tree(args):
    """Recreate tree from lff.txt and dff.txt

    glottolog lff2tree [test]
    """
    try:
        lff.lff2tree(args.repos, args.log)
    except ValueError:  # pragma: no cover
        print("""
Something went wrong! Roll back inconsistent state running

    rm -rf languoids
    git checkout languoids
""")
        raise

    if args.args and args.args[0] == 'test':  # pragma: no cover
        print("""
You can run

    diff -rbB build/tree/ languoids/tree/

to inspect the changes in the directory tree.
""")
    else:
        print("""
Run

    git status

to inspect changes in the directory tree.
You can run

    diff -rbB build/tree/ languoids/tree/

to inspect the changes in detail.

- To discard changes run

    git checkout languoids/tree

- To commit and push changes, run

    git add -A languoids/tree/...

  for any newly created nodes listed under

# Untracked files:
#   (use "git add <file>..." to include in what will be committed)
#
#	languoids/tree/...

  followed by

    git commit -a -m"reason for change of classification"
    git push origin
""")



from __future__ import unicode_literals

import re
import itertools
from datetime import date
import hashlib
from xml.etree import ElementTree

from six import text_type

import attr
from clldutils import iso_639_3
from clldutils.path import read_text, write_text, Path
from csvw import dsv

from .references.bibtex import save

ISO_CODE_PATTERN = re.compile('[a-z]{3}$')
CACHE_DIR = 'iso_639_3_cache'


def read_url(path, cache_dir=None, log=None):
    """
    Delegate scraping to clldutils, since nowadays this requires tweaking the user agent as well.
    """
    if cache_dir:
        cache_dir = Path(cache_dir)
        if log:  # pragma: no cover
            log.debug('retrieving {0} ...'.format(path))
        fpath = cache_dir / hashlib.md5(path.encode('utf8')).hexdigest()
        if not fpath.exists():
            with iso_639_3._open(path) as fp:
                write_text(fpath, fp.read().decode('utf8'))
        else:  # pragma: no cover
            if log:
                log.debug('... from cache {0}'.format(fpath))
        return read_text(fpath)

    with iso_639_3._open(path) as fp:
        return fp.read().decode('utf8')


def valid_iso_code(instance, attr, value):
    if not ISO_CODE_PATTERN.match(value):
        raise ValueError('invalid ISO code in {0}: {1}'.format(attr.name, value))


def normalize_whitespace(s):
    return re.sub('\s+', ' ', s).strip()


@attr.s
class Retirement(object):
    RET_REASON = {  # http://www-01.sil.org/iso639-3/download.asp#retiredDownloads
        'C': 'change',
        'D': 'duplicate',
        'N': 'non-existent',
        'S': 'split',
        'M': 'merge',
    }
    Id = attr.ib(validator=valid_iso_code)
    Ref_Name = attr.ib()
    Ret_Reason = attr.ib(converter=lambda v: Retirement.RET_REASON[v])
    Change_To = attr.ib(
        converter=lambda v: v or None,
        validator=attr.validators.optional(valid_iso_code))
    Ret_Remedy = attr.ib(converter=normalize_whitespace)
    Effective = attr.ib(converter=lambda v: date(*[int(p) for p in v.split('-')]) if v else None)
    cr = attr.ib(default=None)

    @classmethod
    def iter(cls, cache_dir=None, log=None):
        content = read_url(
            'sites/iso639-3/files/downloads/iso-639-3_Retirements.tab',
            cache_dir=cache_dir,
            log=log)
        for d in dsv.reader(content.splitlines(), dicts=True, delimiter='\t'):
            yield cls(**d)


@attr.s
class ChangeRequest(object):
    CHANGE_TYPES = {  # map change types to a sort key
        'Create': 'z',
        'Merge': 'c',
        'Retire': 'a',
        'Split': 'b',
        'Update': 'y'
    }
    Status = attr.ib(
        validator=attr.validators.in_(['Rejected', 'Adopted', 'Pending', 'Partially Adopted']))
    Reference_Name = attr.ib()
    Effective_Date = attr.ib(
        converter=lambda v: date(*[int(p) for p in v.split('-')]) if v else None)
    Change_Type = attr.ib(validator=attr.validators.in_(list(CHANGE_TYPES.keys())))
    Change_Request_Number = attr.ib(converter=lambda v: text_type(v) if v else None)
    Region_Group = attr.ib()
    Affected_Identifier = attr.ib()
    Language_Family_Group = attr.ib()

    @property
    def url(self):
        return iso_639_3.BASE_URL + 'request/' + self.Change_Request_Number

    @property
    def year(self):
        return self.Change_Request_Number.split('-')[0]

    @property
    def pdf(self):
        return '{0}sites/iso639-3/files/change_requests/{1}/{2}.pdf'.format(
            iso_639_3.BASE_URL, self.year, self.Change_Request_Number)

    @classmethod
    def iter(cls, max_year=None, cache_dir=None, log=None):
        path = "code_changes/change_request_index/data/{0}?" \
               "field_change_request_region_grp_tid=All&field_change_request_lf_group_tid=All&" \
               "field_change_instance_chnge_type_tid=All&field_change_request_act_status_tid=All&" \
               "items_per_page=100&page={1}"
        year, page = 2006, 0
        while year < (max_year or date.today().year):
            while True:
                i = 0
                for i, cr in enumerate(
                        list(_iter_tables(
                            read_url(path.format(year, page), cache_dir=cache_dir, log=log)))[0]):
                    yield cls(**{k.replace(' ', '_'): v for k, v in cr.items()})
                if i < 99:
                    break
                page += 1  # pragma: no cover
            year += 1
            page = 0


def change_request_as_source(id_, rows, ref_ids):
    title = "Change Request Number {0}: ".format(id_)
    title += ", ".join(
        "{0} {1} [{2}]".format(r.Status.lower(), r.Change_Type.lower(), r.Affected_Identifier)
        for r in sorted(
            rows,
            key=lambda cr: (ChangeRequest.CHANGE_TYPES[cr.Change_Type], cr.Affected_Identifier)))
    date = None
    for row in rows:
        if row.Effective_Date:
            if date:
                assert date == row.Effective_Date  # pragma: no cover
            else:
                date = row.Effective_Date
    if date:
        title += ' ({0})'.format(date.isoformat())
    fields = {
        'number': id_,
        'title': title,
        'howpublished': rows[0].url,
        'address': "Dallas",
        'author': "ISO 639-3 Registration Authority",
        'publisher': "SIL International",
        'url': rows[0].pdf,
        'year': rows[0].year,
        'hhtype': "overview",
        'lgcode': ', '.join(
            "{0} [{1}]".format(r.Reference_Name, r.Affected_Identifier) for r in rows),
        'src': "iso6393",
    }
    if id_ in ref_ids and ref_ids[id_]:
        fields['glottolog_ref_id'] = ref_ids[id_]
    return id_, ('misc', fields)


def bibtex(api, log, max_year=None):
    """Create a BibTeX file listing records for each past ISO 639-3 change request.

    http://www-01.sil.org/iso639-3/chg_requests.asp?order=CR_Number&chg_status=past
    """
    bib = api.bibfiles['iso6393.bib']
    glottolog_ref_ids = bib.glottolog_ref_id_map

    entries = []

    with api.cache_dir(CACHE_DIR) as cache_dir:
        grouped = itertools.groupby(
            sorted(ChangeRequest.iter(max_year=max_year, cache_dir=cache_dir),
                   key=lambda cr: (cr.Change_Request_Number, cr.Affected_Identifier)),
            lambda cr: cr.Change_Request_Number)
        for id_, rows in grouped:
            entries.append(change_request_as_source(id_, list(rows), glottolog_ref_ids))

    save(entries, bib.fname, None)
    log.info('bibtex written to {0}'.format(bib.fname))
    return len(entries)


def _read_table(table):
    def _text(e):
        if e.find('span') is not None:
            return _text(e.find('span'))
        if e.find('a') is not None:
            return _text(e.find('a'))
        return e.text or ''

    d = ElementTree.fromstring(table)
    header = [e.text.strip() for e in d.findall('.//th')]
    for tr in d.find('tbody').findall('.//tr'):
        yield dict(zip(header, [normalize_whitespace(_text(e)) for e in tr.findall('.//td')]))


def _iter_tables(html):
    start, end = '<table ', '</table>'
    while start in html:
        html = html.split(start, 1)[1]
        table, html = html.split(end, 1)
        yield _read_table(start + table + end)


def code_details(code, cache_dir=None, log=None):
    res = {}
    try:
        for md in _iter_tables(read_url('code/{0}'.format(code), cache_dir=cache_dir, log=log)):
            for row in md:
                for k, v in row.items():
                    if not res.get(k):
                        res[k] = v
    except:  # noqa: E722
        pass
    return res


def get_retirements(max_year=None, cache_dir=None, log=None):
    # retired iso_codes
    rets = list(Retirement.iter(cache_dir=cache_dir, log=log))

    # latest adopted change request affecting each iso_code
    crs = (
        r for r in ChangeRequest.iter(max_year=max_year, cache_dir=cache_dir, log=log)
        if r.Status == 'Adopted')
    crs = sorted(crs, key=lambda r: (r.Affected_Identifier, r.Effective_Date or date.today()))
    crs = itertools.groupby(crs, lambda r: r.Affected_Identifier)
    crs = {id_: list(grp)[-1] for id_, grp in crs}

    # left join
    for ret in rets:
        ret.cr = crs.get(ret.Id)

    # fill Change_To from Ret_Remedy for splits and make it a list for others
    assert all(bool(r.Change_To) == (r.Ret_Reason not in ('split', 'non-existent')) for r in rets)
    assert all(bool(r.Ret_Remedy) == (r.Ret_Reason == 'split') for r in rets)
    iso = re.compile(r'\[([a-z]{3})\]')
    for r in rets:
        if r.Ret_Reason == 'split':
            r.Change_To = iso.findall(r.Ret_Remedy)
        else:
            r.Change_To = [r.Change_To] if r.Change_To else []

    for r in rets:
        if not r.Ret_Remedy:
            r.Ret_Remedy = code_details(r.Id, cache_dir=cache_dir, log=log).get('Retirement Remedy')

    return rets


def retirements(api, log, max_year=None):
    fields = [
        ('Id', 'code'),
        ('Ref_Name', 'name'),
        ('Effective', 'effective'),
        ('Ret_Reason', 'reason'),
        ('Change_To', 'change_to'),
        ('Ret_Remedy', 'remedy'),
    ]
    log.info('read languoid info')
    iso2lang = {l.iso: l for l in api.languoids() if l.iso}
    log.info('retrieve retirement info')
    with api.cache_dir(CACHE_DIR) as cache_dir:
        rets = get_retirements(cache_dir=cache_dir, log=log, max_year=max_year)
    for r in rets:
        lang = iso2lang.get(r.Id)
        if lang is None:
            print('--- Missing retired ISO code: {}'.format(r.Id))
            continue
        for iso in r.Change_To:
            if iso not in iso2lang:
                print('+++ Missing change_to ISO code: {}'.format(iso))
        for f, option in fields:
            lang.cfg.set('iso_retirement', option, getattr(r, f))
        if r.cr and r.cr.Change_Request_Number:
            lang.cfg.set('iso_retirement', 'change_request', r.cr.Change_Request_Number)
        lang.write_info()
    """
    [iso_retirement]
    comment = Interlingue is the later name (currently in use) for this language, created by
        Edgar de Wahl. The [ile] identifier is in ISO 639-2 (as well as ISO 639-3).
        Occidental should be added as another name associated with [ile].
    code = occ
    name = Occidental
    effective = 2007-07-18
    reason = duplicate
    remedy = Merge into Interlingue [ile] as Duplicate
    change_request = 2006-090
    """



from __future__ import unicode_literals

import six

import pytest
import mock

from clldutils.path import copytree

from pyglottolog.__main__ import commands


def _args(api_copy, *args):
    return mock.Mock(repos=api_copy, args=list(args), log=mock.Mock())


def test_show(capsys, api):
    commands.show(_args(api, '**a:key**'))
    assert (b'@misc' if six.PY2 else '@misc') in capsys.readouterr()[0]

    commands.show(_args(api, 'a:key'))
    assert (b'@misc' if six.PY2 else '@misc') in capsys.readouterr()[0]

    commands.show(_args(api, 'abcd1236'))
    assert (b'Classification' if six.PY2 else 'Classificat') in capsys.readouterr()[0]


def test_edit(mocker, api):
    mocker.patch('pyglottolog.commands.subprocess')
    commands.edit(_args(api, 'abcd1236'))


def test_create(capsys, api_copy):
    commands.create(_args(api_copy, 'abcd1234', 'new name', 'language'))
    assert 'Info written' in capsys.readouterr()[0]
    assert 'new name' in [c.name for c in api_copy.languoid('abcd1234').children]


def test_fts(capsys, api_copy):
    with pytest.raises(ValueError):
        commands.refsearch(_args(api_copy, 'Harzani year:1334'))

    commands.refindex(_args(api_copy))
    commands.refsearch(_args(api_copy, 'Harzani year:1334'))
    assert "'Abd-al-'Ali Karang" in capsys.readouterr()[0]

    commands.langindex(_args(api_copy))
    commands.langsearch(_args(api_copy, 'id:abcd*'))
    assert "abcd1234" in capsys.readouterr()[0]

    commands.langsearch(_args(api_copy, 'classification'))
    assert "abcd1234" in capsys.readouterr()[0]


def test_metadata(capsys, api):
    commands.metadata(_args(api))
    assert "longitude" in capsys.readouterr()[0]


def test_lff(capsys, api_copy, encoding='utf-8'):
    commands.tree2lff(_args(api_copy))

    dff = api_copy.build_path('dff.txt')
    dfftxt = dff.read_text(encoding=encoding).replace('dialect', 'Dialect Name')
    dff.write_text(dfftxt, encoding=encoding)
    commands.lff2tree(_args(api_copy))
    assert 'git status' in capsys.readouterr()[0]
    assert api_copy.languoid('abcd1236').name == 'Dialect Name'
    # Old language and dialect names are retained as alternative names:
    assert 'dialect' in api_copy.languoid('abcd1236').names['glottolog']

    commands.tree2lff(_args(api_copy))
    dfftxt = dff.read_text(encoding=encoding)


def test_index(api_copy):
    commands.index(_args(api_copy))
    assert len(list(api_copy.repos.joinpath('languoids').glob('*.md'))) == 8


def test_tree(capsys, api):
    with pytest.raises(commands.ParserError):
        commands.tree(_args(api))

    with pytest.raises(commands.ParserError):
        commands.tree(_args(api, 'xyz'))

    commands.tree(_args(api, 'abc', 'language'))
    out, _ = capsys.readouterr()
    if not isinstance(out, six.text_type):
        out = out.decode('utf-8')
    assert 'language' in out
    assert 'dialect' not in out


def test_languoids(capsys, api, tmpdir):
    commands.languoids(_args(api, '--output={0}'.format(tmpdir), '--version=1.5'))
    out, _ = capsys.readouterr()
    assert '-metadata.json' in out
    assert tmpdir.join('glottolog-languoids-1.5.csv').ensure()


def test_newick(capsys, api):
    commands.newick(_args(api, 'abcd1235'))
    assert 'language' in capsys.readouterr()[0]


def test_htmlmap(api, capsys, tmpdir):
    commands.htmlmap(_args(api, str(tmpdir)), min_langs_for_legend_item=1)
    out, _ = capsys.readouterr()
    assert 'glottolog_map.html' in out


def test_iso2codes(api, tmpdir):
    commands.iso2codes(_args(api, str(tmpdir)))
    assert tmpdir.join('iso2glottocodes.csv').check()


def test_roundtrip(api_copy):
    commands.roundtrip(_args(api_copy, 'a.bib'))


def test_bibfiles_db(api_copy):
    commands.bibfiles_db(_args(api_copy))


def test_check(capsys, api_copy):
    commands.check(_args(api_copy, 'refs'))

    args = _args(api_copy)
    commands.check(args)
    assert 'family' in capsys.readouterr()[0]
    msgs = [a[0] for a, _ in args.log.error.call_args_list]
    assert all('unregistered glottocode' in m for m in msgs)
    assert len(msgs) == 5

    copytree(
        api_copy.tree / 'abcd1234' / 'abcd1235',
        api_copy.tree / 'abcd1235')

    args = _args(api_copy)
    commands.check(args)
    msgs = [a[0] for a, _ in args.log.error.call_args_list]
    assert any('duplicate glottocode' in m for m in msgs)
    assert len(msgs) == 7

    (api_copy.tree / 'abcd1235').rename(api_copy.tree / 'abcd1237')
    args = _args(api_copy)
    commands.check(args)
    msgs = [a[0] for a, _ in args.log.error.call_args_list]
    assert any('duplicate hid' in m for m in msgs)
    assert len(msgs) >= 9


def test_monster(capsys, api_copy):
    commands.bib(_args(api_copy))
    assert capsys.readouterr()[0]



# references

from __future__ import unicode_literals

from .bibfiles import BibFiles, BibFile, Entry
from .hhtypes import HHTypes
from .isbns import Isbns, Isbn
from .roman import introman, romanint

__all__ = [
    'BibFiles', 'BibFile', 'Entry',
    'Isbns', 'Isbn',
    'HHTypes',
    'introman', 'romanint',
]



# isbns.py

from __future__ import unicode_literals

from six.moves import map

import re

__all__ = ['Isbns', 'Isbn']


class Isbns(list):

    class Parser(object):
        @staticmethod
        def parse(ma):
            return ''.join(g for g in ma.groups() if g is not None)

    class Numeric(Parser):
        pattern = re.compile(r'(97[89])?(\d{9})([\dXx])(?![\dXx])')

    class Hyphened(Parser):
        pattern = re.compile(r'''
            (?:
              (?:ISBN[- ])? (97[89])-
              |
              ISBN-10[ ]
            )?
            (\d{1,5})- (\d+)- (\d+)- ([\dXx])(?![\dXx])''', flags=re.VERBOSE)

    class Ten99(Parser):
        pattern = re.compile(r'(99)-(\d{7})-([\dXx])(?![\dXx])')

    @classmethod
    def _iterparse(cls, s, pos=0, parsers=(Numeric, Hyphened, Ten99), delimiters=(', ', ' ')):
        while True:
            for p in parsers:
                ma = p.pattern.match(s, pos)
                if ma is not None:
                    yield p.parse(ma)
                    pos += len(ma.group())
                    break
            else:
                raise ValueError('no matching ISBN pattern at index %s: %r' % (pos, s))
            if pos == len(s):
                return
            for delim in delimiters:
                if s.startswith(delim, pos):
                    pos += len(delim)
                    break
            else:
                raise ValueError('no matching delimiter at index %s: %r' % (pos, s))

    @classmethod
    def from_field(cls, field):
        isbns = map(Isbn, cls._iterparse(field))
        seen = set()
        return cls(i for i in isbns if i not in seen and not seen.add(i))

    def to_string(self):
        return ', '.join(i.digits for i in self)


class Isbn(object):
    """A 13 digit ISBN from a string of 13 or 10 digits.

    see also https://en.wikipedia.org/wiki/International_Standard_Book_Number
    """

    _isbn10_prefix = '978'

    @staticmethod
    def _isbn13_check_digit(digits):
        assert len(digits) in (12, 13)
        halfes = (digits[i:12:2] for i in (0, 1))
        odd, even = (sum(map(int, h)) for h in halfes)
        return str(-(odd + 3 * even) % 10)

    @staticmethod
    def _isbn10_check_digit(digits):
        assert len(digits) in (9, 10)
        result = sum(i * int(d) for i, d in enumerate(digits[:9], 1)) % 11
        return 'X' if result == 10 else str(result)

    def __init__(self, digits):
        if len(digits) == 13:
            check = self._isbn13_check_digit(digits)
        elif len(digits) == 10:
            digits = digits.upper()
            check = self._isbn10_check_digit(digits)
        else:
            raise ValueError('invalid ISBN digits length (%s): %r' % (len(digits), digits))

        if digits[-1] != check:
            raise ValueError('invalid ISBN check digit (%s instead of %s): %r'
                             % (digits[-1], check, digits))

        if len(digits) == 10:  # convert
            digits = self._isbn10_prefix + digits[:9]
            digits += self._isbn13_check_digit(digits)
        self.digits = digits

    def __hash__(self):
        return hash(self.digits)

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.digits == other.digits
        return NotImplemented  # pragma: no cover

    def __ne__(self, other):
        if isinstance(other, self.__class__):
            return self.digits != other.digits
        return NotImplemented  # pragma: no cover

    def __repr__(self):
        return '%s(%r)' % (self.__class__.__name__, self.digits)



from __future__ import unicode_literals

import pytest

from pyglottolog.references.libmonster import (markconservative, markall,
    add_inlg_e, INLG, keyid, pyear, pagecount, lgcode, grp2fd)


def test_markconcservative(tmpdir, hhtypes):
    res = markconservative(
        {1: ('article', {'title': 'Grammar'})},
        hhtypes.triggers,
        {1: ('article', {'title': 'Grammar'})},
        hhtypes,
        str(tmpdir / 'marks.txt'),
        verbose=False)
    assert res[1][1]['hhtype'].split()[0] == 'grammar'

    # If a higher hhtype is computed, this cancels out previous computations.
    res = markconservative(
        {1: ('article', {'title': 'grammar', 'lgcode': 'abc'})},
        hhtypes.triggers,
        {1: ('article', {'title': 'other', 'hhtype': 'other', 'lgcode': 'abc'})},
        hhtypes,
        str(tmpdir / 'marks.txt'),
        verbose=False)
    assert 'hhtype' not in res[1][1]


def test_markall(api, hhtypes):
    bib = {
        1: ('article', {'title': "other grammar of lang"}),
        2: ('article', {'title': "grammar of lang and dial"}),
        3: ('article', {'title': "other"}),
        4: ('article', {'title': "grammar and phonologie and morphologie"})
    }
    markall(bib, hhtypes.triggers, verbose=False, rank=lambda l: hhtypes[l].rank)
    assert 'grammar' in bib[1][1]['hhtype']
    assert 'morphologie and phonologie;grammar' in bib[4][1]['hhtype']

    markall(bib, api.triggers['lgcode'], verbose=False)
    assert 'language' in bib[1][1]['lgcode']


def test_add_inlg_e(api):
    res = add_inlg_e(
        {1: ('article', {'title': 'Grammar of language'})},
        api.triggers[INLG],
        verbose=False)
    assert res[1][1][INLG] == 'language [abc]'


@pytest.mark.parametrize('fields, expected', [
    ({}, '__missingcontrib__'),
    ({'author': 'An Author'}, 'author_no-titlend'),
    ({'editor': 'An Author'}, 'author_no-titlend'),
    ({'author': 'An Author', 'title': 'A rather long title'}, 'author_rather-longnd'),
    ({'author': 'An Author', 'title': 'Title', 'year': '2014'}, 'author_title2014'),
    ({'author': 'An Author', 'volume': 'IV'}, 'author_no-titleivnd'),
    ({'author': 'An Author', 'extra_hash': 'a'}, 'author_no-titlenda'),
])
def test_keyid(fields, expected):
    assert keyid(fields, {}) == expected


def test_keyid_invalid(capsys):
    assert keyid({'author': 'An Author and '}, {}) == 'author_no-titlend'
    assert 'Unparsed' in capsys.readouterr()[0]


@pytest.mark.parametrize('year, expected', [
    ('', '[nd]'),
    ('1931', '1931'),
    ('1931-32', '1931-1932'),
])
def test_pyear(year, expected):
    assert pyear(year) == expected


@pytest.mark.parametrize('pages, expected', [
    ('', ''),
    ('1', '1'),
    ('10-20', '11'),
    ('10-20,v-viii', '4+11'),
    ('20,viii', '8+20'),
    ('10-2', '3'),  # interpreted as 10-12
])
def test_pagecount(pages, expected):
    assert pagecount(pages) == expected


@pytest.mark.parametrize('lgcode_, expected', [
    ('', []),
    ('[abc]', ['abc']),
    ('abc,NOCODE_Abc', ['abc', 'NOCODE_Abc']),
])
def test_lgcode(lgcode_, expected):
    assert lgcode((None, {'lgcode': lgcode_})) == expected


@pytest.mark.parametrize('input_, expected', [
    ([(1, 2), (1, 3), (2, 4), (1, 5)], {1: {2: 1, 3: 1, 5: 1}, 2: {4: 1}}),
])
def test_grp2fd(input_, expected):
    assert grp2fd(input_) == expected



from __future__ import unicode_literals

import pytest

from pyglottolog import languoids


def test_legacy_import():
    from pyglottolog import api
    from pyglottolog import Glottolog
    assert api.Glottolog is Glottolog


def test_glottolog_invalid_repos(tmpdir):
    from pyglottolog import Glottolog
    with pytest.raises(ValueError, match=r'missing tree dir'):
        Glottolog(str(tmpdir))


def test_paths(api):
    assert api.ftsindex


def test_languoid(api):
    assert api.languoid('abc').name == 'language'


def test_languoids(api):
    assert len(list(api.languoids())) == 5
    assert len(list(api.languoids(maxlevel=languoids.Level.family))) == 1
    assert len(list(api.languoids(maxlevel=languoids.Level.language))) == 3
    assert len(api.languoids_by_code()) == 8
    assert 'NOCODE_Family-name' in api.languoids_by_code()


def test_newick_tree(api):
    assert api.newick_tree(start='abcd1235') == \
           "('dialect [abcd1236]':1)'language [abcd1235][abc]-l-':1;"
    assert api.newick_tree(start='abcd1235', template='{l.id}') == "(abcd1236:1)abcd1235:1;"
    assert set(api.newick_tree().split('\n')) == {
        "(('isolate {dialect} [dial1234]':1)'isolate [isol1234]-l-':1)'isolate [isol1234]':1;",
        "(('dialect [abcd1236]':1)'language [abcd1235][abc]-l-':1)'family [abcd1234][aaa]':1;"
    }


def test_hhtypes(api):
    assert len(api.hhtypes) == 2


def test_load_triggers(api):
    assert len(api.triggers) == 2


def test_macroarea_map(api):
    assert api.macroarea_map['abc'] == 'Eurasia'



from __future__ import unicode_literals

import pytest

from pyglottolog.languoids import (Languoid, EndangermentStatus,
    Glottocodes, Glottocode, Level, Country, Reference, Macroarea,
    ClassificationComment, EthnologueComment)


def test_legacy_imports():
    from pyglottolog import objects
    assert objects.Glottocode is Glottocode
    assert objects.Macroarea is Macroarea
    assert objects.Level is Level
    assert objects.Reference is Reference


def test_Glottocodes(tmpdir):
    json = tmpdir / 'glottocodes.json'
    json.write_text('{}', encoding='ascii')

    glottocodes = Glottocodes(str(json))
    gc = glottocodes.new('a', dry_run=True)
    assert gc.startswith('aaaa')
    assert gc not in glottocodes
    gc = glottocodes.new('a')
    assert gc in glottocodes
    # make sure it's also written to file:
    assert gc in Glottocodes(str(json))
    assert len(list(Glottocodes(str(json)))) == 1


def test_es():
    assert EndangermentStatus.get('nearly extinct') == EndangermentStatus.critical


def test_EndangermentStatus():
    c = EndangermentStatus.critical
    assert EndangermentStatus.get(c) == c

    with pytest.raises(ValueError):
        EndangermentStatus.get(123)


@pytest.mark.parametrize('input_, valid', [
    ('abcd1234', True),
    ('a12d3456', True),
    ('abcd123', False),
    ('12d3456', False),
    ('aNOCODE', False),
    ('NOCODE_abd', False),
    ('nocode', False),
])
def test_pattern(input_, valid, _match=Glottocode.pattern.match):
    assert (_match(input_) is not None) == valid


def test_Country():
    assert Country.from_text('Germany').id == 'DE'
    assert Country.from_name('abcdefg') is None
    assert Country.from_id('abcdefg') is None


def test_Glottocode_validation():
    with pytest.raises(ValueError):
        Glottocode('a2')


def test_Glottocode_ordering():
    assert sorted([Glottocode('abcd1235'), Glottocode('abcd1234')])[0] == Glottocode('abcd1234')
    assert Glottocode('zzzz9999') > Glottocode('abcd1234')
    assert Glottocode('abcd1234') <= Glottocode('abcd1234')


def test_Reference():
    ref = Reference('bib:key', '12-34', 'German')
    assert '{0}'.format(ref) == '**bib:key**:12-34<trigger "German">'
    Reference.from_list(['{0}'.format(ref)])

    with pytest.raises(ValueError):
        Reference.from_list(['abc'])


def test_ClassificationComment(mocker):
    cc = ClassificationComment(family='**bib:key**')
    log = mocker.Mock()
    cc.check(mocker.Mock(), [], log)
    assert log.error.called
    log = mocker.Mock()
    cc.check(mocker.Mock(), ['bib:key'], log)
    assert not log.error.called


def test_EthnologueComment(mocker):
    with pytest.raises(ValueError):
        EthnologueComment('abc', 't')

    with pytest.raises(ValueError):
        EthnologueComment('abc', 'missing', 'E15')

    with pytest.raises(ValueError):
        EthnologueComment('abc', 'missing', 'E16')

    with pytest.raises(ValueError):
        EthnologueComment('abc', 'missing', 'E16', '\u00e4\u00f6\u00fc'.encode('utf-8'))

    log = mocker.Mock()
    ec = EthnologueComment('abc', 'missing', 'E16', 'abc')
    ec.check(mocker.Mock(), [], log)
    assert not log.error.called

    log = mocker.Mock()
    ec = EthnologueComment('abc', 'missing', 'E16', '**bib:key**')
    ec.check(mocker.Mock(), [], log)
    assert log.error.called


def test_Level(api):
    assert Level.dialect > Level.language
    assert Level.language == api.languoid('abcd1235').level
    with pytest.raises(ValueError):
        Level.get('abcde')


def test_Languoid_sorting(api):
    assert api.languoid('abcd1235') < api.languoid('abcd1236')
    assert api.languoid('abcd1236') >= api.languoid('abcd1235')


def test_factory(tmpdir, api_copy):
    f = Languoid.from_dir(api_copy.tree / 'abcd1234')
    assert f.category == 'Family'
    l = Languoid.from_dir(api_copy.tree / f.id / 'abcd1235')
    assert l.name == 'language'
    assert 'abcd1235' in repr(l)
    assert 'language' in '%s' % l
    assert l.level == Level.language
    assert l.latitude == pytest.approx(0.5)
    assert l.longitude == pytest.approx(0.5)
    l.latitude, l.longitude = 1.0, 1.0
    assert l.latitude == pytest.approx(1.0)
    assert l.longitude == pytest.approx(1.0)
    assert l.iso_code == 'abc'
    l.iso_code = 'cde'
    assert l.iso == 'cde'
    assert l.hid == 'abc'
    l.hid = 'abo'
    assert l.hid == 'abo'
    assert l.id == 'abcd1235'

    assert len(l.macroareas) == 2
    l.macroareas = [Macroarea.africa]
    assert l.macroareas == [Macroarea.africa]

    l.countries = api_copy.countries[:2]
    assert len(l.countries) == 2

    assert l.parent == f
    assert f.children[0] == l
    assert l.children[0].family == f
    l.write_info(str(tmpdir))
    assert (tmpdir / 'abcd1235').exists()
    assert isinstance(api_copy.languoid('abcd1235').iso_retirement.asdict(), dict)
    assert l.classification_comment is None
    l.endangerment = 'nearly extinct'
    assert l.endangerment == EndangermentStatus.critical
    assert l.names == {}
    l.cfg['altnames'] = {'glottolog': 'xyz'}
    assert 'glottolog' in l.names
    assert l.identifier == {}
    l.cfg['identifier'] = {'multitree': 'xyz'}
    assert 'multitree' in l.identifier


def test_isolate(api):
    l = Languoid.from_dir(api.tree / 'isol1234')
    assert l.isolate
    assert l.parent is None
    assert l.family is None


def test_attrs(api):
    l = Languoid.from_name_id_level(
        api.tree, 'name', 'abcd1235', Level.language, hid='NOCODE')
    l.name = 'other'
    assert l.name == 'other'
    with pytest.raises(AttributeError):
        l.glottocode = 'x'
    with pytest.raises(AttributeError):
        l.id = 'x'
    assert l.id == l.glottocode
    assert l.hid == 'NOCODE'



from __future__ import unicode_literals

import six

import pytest

from pyglottolog.references.bibfiles_db import Database, distance


def test_Database(capsys, tmpdir, bibfiles_copy):
    fpath = str(tmpdir / 'test.sqlite3')
    bibfiles_copy.to_sqlite(fpath)
    assert 'ENTRYTYPE' in capsys.readouterr()[0]

    bibfiles_copy.to_sqlite(fpath)
    bibfiles_copy.to_sqlite(fpath, rebuild=True)
    capsys.readouterr()

    db = Database(fpath)
    db.is_uptodate(bibfiles_copy[1:], verbose=True)
    assert len(capsys.readouterr()[0].splitlines()) == 3

    db.to_bibfile(str(tmpdir / 'out.bib'))
    db.to_csvfile(str(tmpdir / 'out.csv'))
    db.to_replacements(str(tmpdir / 'out.json'))

    db.trickle(bibfiles_copy)
    assert '2 changed 1 added in a' in capsys.readouterr()[0]

    key, (entrytype, fields) = db[('b.bib', 'arakawa97')]
    assert entrytype == 'article'
    assert fields['volume'] == '16'

    db.stats()

    db.show_splits()

    db.show_merges()

    db.show_identified()

    db.show_combined()


d1 = {
    'ENTRYTYPE': 'article',
    'author': 'An Author',
    'year': '1998',
    'title': 'The Title',
}


@pytest.mark.parametrize('left, right, expected', [
    ({}, {}, 0.0),
    (d1, d1, 0.0),
    (d1, {}, 1.0),
    ({}, d1, 1.0),
    (d1, {'author': 'An Author'}, 0.0),
    (d1, {'author': 'Another Author'}, pytest.approx(0.2173, rel=.001)),
    (d1, {'author': 'An Author', 'title': 'Another Title'}, pytest.approx(0.13636, rel=.001)),
])
def test_distance(left, right, expected):
    assert distance(left, right) == expected



# bibtex.py - bibtex file parsing/serialization

# TODO: make check fail on non-whitespace between entries (bibtex 'comments')

import io
import functools
import collections
import unicodedata

from six import PY2, text_type

from pybtex.database.input.bibtex import LowLevelParser, Parser, UndefinedMacro
from pybtex.scanner import PybtexSyntaxError
from pybtex.exceptions import PybtexError
from pybtex.textutils import whitespace_re
from pybtex.bibtex.utils import split_name_list
from pybtex.database import Person

from clldutils.path import as_posix, memorymapped

FIELDORDER = [
    'author', 'editor', 'title', 'booktitle', 'journal',
    'school', 'publisher', 'address',
    'series', 'volume', 'number', 'pages', 'year', 'issn', 'url',
]


def load(filename, preserve_order=False, encoding=None):
    cls = collections.OrderedDict if preserve_order else dict
    return cls(iterentries(filename, encoding))


def identity(x):
    return x


def iterentries_from_text(text, encoding='utf-8'):
    py2_decode = identity
    if hasattr(text, 'read'):
        text = text.read(-1)
    if not isinstance(text, text_type):
        text = text.decode(encoding)
    for entrytype, (bibkey, fields) in LowLevelParser(text):
        fields = {
            py2_decode(name).lower():
                whitespace_re.sub(' ', py2_decode(''.join(values)).strip())
            for name, values in fields}
        yield py2_decode(bibkey), (py2_decode(entrytype), fields)


def iterentries(filename, encoding=None):
    encoding = encoding or 'utf8'
    with memorymapped(as_posix(filename)) as source:
        try:
            for entrytype, (bibkey, fields) in iterentries_from_text(source, encoding):
                yield entrytype, (bibkey, fields)
        except PybtexSyntaxError as e:  # pragma: no cover
            debug_pybtex(source, e)


def debug_pybtex(source, e):  # pragma: no cover
    start, line, pos = e.error_context_info
    print('BIBTEX ERROR on line %d, last parsed lines:' % line)
    print(source[start:start + 500] + '...')
    raise e


def names(s):
    for name in split_name_list(s):
        try:
            yield Name.from_string(name)
        except PybtexError as e:  # pragma: no cover
            print(repr(e))


class Name(collections.namedtuple('Name', 'prelast last given lineage')):

    __slots__ = ()

    @classmethod
    def from_string(cls, name):
        person = Person(name)
        ntypes = ('prelast_names', 'last_names', 'first_names', 'middle_names', 'lineage_names')
        prelast, last, first, middle, lineage = (' '.join(getattr(person, part)) for part in ntypes)
        given = ' '.join(n for n in (first, middle) if n)
        return cls(prelast, last, given, lineage)


def save(entries, filename, sortkey, encoding='utf-8', normalize='NFC'):
    with io.open(as_posix(filename), 'w', encoding=encoding, errors='strict') as fd:
        dump(entries, fd, sortkey, normalize)


def dump(entries, fd, sortkey=None, normalize='NFC'):
    assert sortkey in [None, 'bibkey']
    if sortkey is None:
        if isinstance(entries, collections.OrderedDict):  # pragma: no cover
            items = entries.items()
        elif isinstance(entries, dict):  # pragma: no cover
            raise ValueError('dump needs sortkey or ordered entries')
        else:
            items = entries
    else:  # elif sortkey == 'bibkey':
        items = (
            (bibkey, entries[bibkey])
            for bibkey in sorted(entries, key=lambda bibkey: bibkey.lower()))
    """Reserved characters (* -> en-/decoded by latexcodec)
    * #: \#
      $: \$
      %: \%
      ^: \^{} \textasciicircum
    * &: \&
    * _: \_
      {: \{
      }: \}
      ~: \~{} \textasciitilde
      \: \textbackslash{}
      <: \textless
      >: \textgreater
    """
    assert normalize in (None, '', 'NFC', 'NFKC', 'NFD', 'NFKD')
    if normalize:
        normalize = functools.partial(unicodedata.normalize, normalize)
    else:  # pragma: no cover
        normalize = identity
    fd.write(u'# -*- coding: utf-8 -*-\n')
    for bibkey, (entrytype, fields) in items:
        fd.write(u'@%s{%s' % (entrytype, bibkey))
        for k, v in fieldorder.itersorted(fields):
            fd.write(u',\n    %s = {%s}' % (k, normalize(v.strip())))
        fd.write(u'\n}\n' if fields else u',\n}\n')


class Ordering(dict):
    """Key order for iterating over dicts (unknown keys last alphabetic)."""

    _missing = float('inf')

    @classmethod
    def fromlist(cls, keys):
        """Define the order of keys as given."""
        return cls((k, i) for i, k in enumerate(keys))

    def itersorted(self, dct):
        """Iterate over dct (key, value) pairs in the defined order."""
        for key in sorted(dct, key=self._itersorted_key):
            yield key, dct[key]

    def _itersorted_key(self, key):
        return self[key], key

    def __missing__(self, key):
        return self._missing


fieldorder = Ordering.fromlist(FIELDORDER)


def check(filename, encoding=None):
    parser = CheckParser(encoding=encoding)
    parser.parse_file(as_posix(filename))
    return parser.error_count


class CheckParser(Parser):
    """Unline LowLevelParser also parses names, macros, etc."""

    def __init__(self, *args, **kwargs):
        super(CheckParser, self).__init__(*args, **kwargs)
        self.error_count = 0

    def handle_error(self, error):  # pragma: no cover
        print('%r' % error)
        self.error_count += 1
        if not isinstance(error, UndefinedMacro):
            raise error

    def process_entry(self, *args, **kwargs):
        try:
            super(CheckParser, self).process_entry(*args, **kwargs)
        except PybtexError as e:  # pragma: no cover
            print(e)
            self.error_count += 1



from setuptools import setup, find_packages


setup(
    name='pyglottolog',
    version='1.4.0',
    author='Robert Forkel',
    author_email='forkel@shh.mpg.de',
    description='python package for glottolog data curation',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    keywords='data linguistics',
    license='Apache 2.0',
    url='https://github.com/clld/pyglottolog',
    packages=find_packages(where='src'),
    package_dir={'': 'src'},
    platforms='any',
    python_requires='>=2.7,!=3.0.*,!=3.1.*,!=3.2.*,!=3.3.*',
    include_package_data=True,
    zip_safe=False,
    entry_points={
        'console_scripts': ['glottolog=pyglottolog.__main__:main'],
    },
    install_requires=[
        'six>=1.9',
        'csvw',
        'clldutils>=2.5.2',
        'pycldf',
        'sqlalchemy',
        'tqdm',
        'pybtex>=0.22',
        'latexcodec',
        'unidecode',
        'whoosh',
        'attrs>=18.1',
        'pycountry>=18.12.8',
        'termcolor',
        'newick',
        'markdown',
        'bs4',
        'requests',
    ],
    extras_require={
        'dev': ['tox>=2.9', 'flake8', 'pep8-naming', 'wheel', 'twine'],
        'test': ['mock>=2', 'pytest>=3.6', 'pytest-mock', 'pytest-cov'],
    },
    classifiers=[
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],
)



# coding: utf8
from __future__ import unicode_literals, print_function, division
import re
from collections import OrderedDict

import requests
from bs4 import BeautifulSoup

from clldutils.misc import nfilter
from clldutils import jsonlib

BASE_URL = "http://www.endangeredlanguages.com"
STORE = 'endangeredlanguages.json'


def read_store(fname):  # pragma: no cover
    return jsonlib.load(fname) if fname.exists() else {}


def store(details_, fname):  # pragma: no cover
    db = read_store(fname)
    if not details_:
        return db
    db[details_['id']] = details_
    ordered = OrderedDict()
    for k in sorted(list(db.keys()), key=lambda lid: int(lid)):
        v = OrderedDict()
        for key in sorted(list(db[k].keys())):
            if key != 'id':
                v[key] = db[k][key]
        ordered[k] = v
    jsonlib.dump(ordered, fname, indent=4)
    return db


def get_soup(path):  # pragma: no cover
    print('... fetch {0}'.format(path))
    return BeautifulSoup(requests.get(BASE_URL + path).content, 'html.parser')


def details(path):  # pragma: no cover
    soup = get_soup(path)
    if not soup.find('h2'):
        return
    res = dict(id=path.split('/')[-1], name=soup.find('h2').get_text())
    data = OrderedDict()
    for tr in soup.find_all('tr'):
        tds = list(tr.find_all('td'))
        if len(tds) == 3:
            data[tds[0].get_text().strip()] = tds[2].get_text().strip()

    names = data.get('ALSO KNOWN AS')
    if names:
        res['alternative_names'] = nfilter([n.strip() for n in names.split(',')])
    if data.get('CODE AUTHORITY') == 'ISO 639-3':
        res['iso-639-3'] = data.get('LANGUAGE CODE')
    return res


def scrape(api, update=False):  # pragma: no cover
    store_path = api.repos.joinpath('links', STORE)
    db = read_store(store_path)
    lang_url = re.compile('/lang/(?P<id>[0-9]+)$')
    done = set()
    for a in get_soup('/lang/region').find_all('a', href=True):
        if a['href'].startswith('/lang/country/') and (a['href'] not in done):
            for _a in get_soup(a['href']).find_all('a', href=True):
                match = lang_url.match(_a['href'])
                if match:
                    lid = match.group('id')
                    if lid not in db or update:
                        db = store(details(_a['href']), store_path)
            done.add(a['href'])
            print(len(done))



from __future__ import unicode_literals

import pytest

from pyglottolog.references.roman import introman, romanint


def test_legacy_import():
    from pyglottolog.monsterlib import roman
    assert roman.romanint is romanint


@pytest.mark.parametrize('input_, expected', [
    (5, 'v',),
    (8, 'viii'),
])
def test_introman(input_, expected):
    assert introman(input_) == expected


def test_roundtrip(n=2000):
    for i in range(1, n + 1):
        assert romanint(introman(i)) == i
